import collections

import numpy as np

from histomicstk.preprocessing import color_conversion
from histomicstk.utils import sample_pixels


def reinhard_stats(
        slide_path,
        sample_fraction,
        magnification=None,
        tissue_seg_mag=1.25,
        invert_image=False,
        style=None,
        frame=None,
        default_img_inversion=False):
    """Samples a whole-slide-image to determine colorspace statistics (mean,
    variance) needed to perform global Reinhard color normalization.

    Normalizing individual tiles independently creates a significant bias
    in the results of segmentation and feature extraction, as the color
    statistics of each tile in a whole-slide image can vary significantly.
    To remedy this, we sample a subset of pixels from the entire whole-slide
    image in order to estimate the global mean and standard deviation of
    each channel in the Lab color space that are needed for reinhard color
    normalization.

    Parameters
    ----------
    slide_path : str
        path and filename of slide.
    sample_fraction : double
       Fraction of pixels to sample (range (0, 1]).
    magnification : scalar
        Desired magnification for sampling. Defaults to native scan
        magnification.
    tissue_seg_mag: double, optional
        low resolution magnification at which foreground will be segmented.
        Default value = 1.25.

    Returns
    -------
    Mu : array_like
        A 3-element array containing the means of the target image channels
        in sample_pix_lab color space.
    Sigma : array_like
        A 3-element list containing the standard deviations of the target image
        channels in sample_pix_lab color space.

    Notes
    -----
    Returns a namedtuple.

    See Also
    --------
    histomicstk.preprocessing.color_conversion.lab_mean_std
    histomicstk.preprocessing.color_normalization.reinhard

    """
    # generate a sampling of sample_pixels_rgb pixels from whole-slide image
    sample_pixels_rgb = sample_pixels(
        slide_path,
        sample_fraction=sample_fraction,
        magnification=magnification,
        tissue_seg_mag=tissue_seg_mag,
        invert_image=invert_image,
        style=style,
        frame=frame,
        default_img_inversion=default_img_inversion,
    )

    # reshape the Nx3 pixel array into a 1 x N x 3 image for lab_mean_std
    sample_pixels_rgb = np.reshape(sample_pixels_rgb,
                                   (1, sample_pixels_rgb.shape[0], 3))

    # compute mean and stddev of sample pixels in Lab space
    Mu, Sigma = color_conversion.lab_mean_std(sample_pixels_rgb)

    # build named tuple for output
    ReinhardStats = collections.namedtuple('ReinhardStats', ['Mu', 'Sigma'])
    stats = ReinhardStats(Mu, Sigma)

    return stats


def reinhard_stats_rgb(rgb_image):
    rgb_pixels = np.reshape(rgb_image, (1, rgb_image.shape[0] * rgb_image.shape[1], 3))

    Mu, Sigma = color_conversion.lab_mean_std(rgb_pixels)

    # build named tuple for output
    return collections.namedtuple('ReinhardStats', ['Mu', 'Sigma'])(Mu, Sigma)



import numpy as np


def chan_vese(im_input, im_mask, sigma,
              dt=1.0, mu=0.2, lambda1=1, lambda2=1, iter=100):
    """Region-based level sets.

    Region-based level set implementation based on the Chan-Vese method.
    Provides cost terms for boundary length, the variance of intensities inside
    the zero level-set, and the variance of external intensities. Robust to
    initialization.

    Parameters
    ----------
    im_input : array_like
        A floating-point intensity image.
    im_mask : array_like
        A binary mask initializing the level set image. Regions corresponding
        to the interior of the level-set function have value 1, with other
        regions having value 0.
    sigma : double
        Standard deviation of smoothing filter for input image im_input.
    dt : double
        Time step for evolving the level-set function im_phi. Default value = 1.0.
    mu : double
        Boundary length weight for energy function. Default value = 0.2.
    lambda1 : double
        Internal variance weight for energy function. Default value = 1.
    lambda2 : double
        External variance weight for energy function. Default value = 1.
    iter : double
        Number of iterations to evolve curve level set function over. Default
        value = 100.


    Returns
    -------
    im_phi : array_like
        An intensity image where the zero level set defines object boundaries.
        Can be further processed with fast marching methods or other to obtain
        smooth boundaries, or simply thresholded to define the object mask.

    See Also
    --------
    histomicstk.segmentation.nuclear.gaussian_voting

    References
    ----------
    .. [#] C. Li, C. Xu, C. Gui, M.D. Fox, "Distance Regularized Level Set
       Evolution and Its Application to Image Segmentation," in IEEE
       Transactions on Image Processing, vol.19,no.12,pp.3243-54, 2010.

    """
    import scipy.ndimage as ndi

    # smoothed gradient of input image
    im_input = ndi.gaussian_filter(im_input, sigma, mode='constant', cval=0)

    # generate signed distance map
    im_phi = mask_to_sdf(im_mask)

    # evolve level set function
    for _i in range(iter):

        # calculate interior and exterior averages
        C1 = np.sum(im_input[im_phi > 0]) / (np.sum(im_phi > 0) + 1e-10)
        C2 = np.sum(im_input[im_phi <= 0]) / (np.sum(im_phi <= 0) + 1e-10)
        Force = lambda2 * (im_input - C2) ** 2 - lambda1 * (im_input - C1) ** 2

        # curvature of image
        Curvature = kappa(im_phi)

        # evolve
        im_phi += dt * Force / np.max(np.abs(Force)) + mu * Curvature

    return im_phi


def mask_to_sdf(im_mask):
    from scipy.ndimage import distance_transform_edt as dtx

    # convert binary mask to signed distance function

    Phi0 = dtx(1 - im_mask) - dtx(im_mask) + im_mask - 1 / 2
    return Phi0


def kappa(im_phi):

    dPhi = np.gradient(im_phi)  # calculate gradient of level set image
    xdPhi = np.gradient(dPhi[1])
    ydPhi = np.gradient(dPhi[0])
    K = (xdPhi[1] * (dPhi[0]**2) - 2 * xdPhi[0] * dPhi[0] * dPhi[1] +
         ydPhi[0] * (dPhi[1]**2)) / ((dPhi[0]**2 + dPhi[1]**2 + 1e-10)**(3 / 2))
    K *= (xdPhi[1]**2 + ydPhi[0]**2)**(1 / 2)
    return K



from .sda_to_rgb import sda_to_rgb


def od_to_rgb(im_od):
    """Transforms input optical density image `im_od` into RGB space

    Parameters
    ----------
    im_od : array_like
        A floating-point image of optical density values obtained
        from rgb_to_od.

    Returns
    -------
    im_rgb : array_like
        A floating-point multi-channel image with intensity
        values in the range [0, 255].

    See Also
    --------
    histomicstk.preprocessing.color_conversion.rgb_to_od,
    histomicstk.preprocessing.color_conversion.sda_to_rgb,
    histomicstk.preprocessing.color_deconvolution.color_deconvolution,
    histomicstk.preprocessing.color_deconvolution.color_convolution

    """
    return sda_to_rgb(im_od, None)  # compatibility mode



"""
Created on Sat Aug 24 16:01:26 2019.

@author: tageldim
"""

import numpy as np
from shapely.geometry.polygon import Polygon
from shapely.ops import unary_union

from histomicstk.annotations_and_masks.masks_to_annotations_handler import \
    _parse_annot_coords
from histomicstk.annotations_and_masks.pyrtree.rtree import Rect, RTree
from histomicstk.utils.general_utils import Base_HTK_Class


class Polygon_merger_v2(Base_HTK_Class):
    """Methods to merge contiguous polygons from whole-slide image."""

    def __init__(self, contours_df, **kwargs):
        """Init Polygon_merger object.

        Arguments:
        ---------
        contours_df : pandas DataFrame
            The following columns are needed.

            group : str
                annotation group (ground truth label).
            ymin : int
                minimum y coordinate
            ymax : int
                maximum y coordinate
            xmin : int
                minimum x coordinate
            xmax : int
                maximum x coordinate
            coords_x : str
                vertex x coordinates comma-separated values
            coords_y
                vertex y coordinated comma-separated values
        merge_thresh : int
            how close do the polygons need to be (in pixels) to be merged
        verbose : int
            0 - Do not print to screen
            1 - Print only key messages
            2 - Print everything to screen
        monitorPrefix : str
            text to prepend to printed statements

        """
        from pandas import DataFrame

        # see: https://stackoverflow.com/questions/8187082/how-can-you-set-...
        # class-attributes-from-variable-arguments-kwargs-in-python
        default_attr = {
            'verbose': 1,
            'monitorPrefix': '',
            'merge_thresh': 3,
        }
        default_attr.update(kwargs)
        super().__init__(default_attr=default_attr)

        # This is where contours will be stored
        self.contours_df = contours_df
        self.contours_df.reset_index(inplace=True, drop=True)
        self.new_contours = DataFrame(columns=self.contours_df.columns)

        # prepwork
        self.buffer_size = self.merge_thresh + 3
        self.unique_groups = set(self.contours_df.loc[:, 'group'])

    def set_contours_slice(self, group):
        """Slice a single group from self.contours_df."""
        self.contours_slice = self.contours_df.loc[
            self.contours_df.loc[:, 'group'] == group, :]

    def create_rtree(self):
        """Add contour bounding boxes to R-tree."""
        self.rtree = RTree()
        for cidx, cont in self.contours_slice.iterrows():
            self.rtree.insert('polygon-%d' % cidx, Rect(
                minx=cont['xmin'], miny=cont['ymin'],
                maxx=cont['xmax'], maxy=cont['ymax']))

    def set_tree_dict(self):
        """Get tree in convenience dict format (dicts inside dicts)."""
        def _traverse(node):
            """Recursively traverse tree till you get to leafs."""
            if not node.is_leaf():
                node_dict = {}
                for c in node.children():
                    node_dict[c.index] = _traverse(c)
                return node_dict
            else:
                return node.index

        self.tree_dict = _traverse(self.rtree.cursor)

    def set_hierarchy(self):
        """Get hierarchy of node indices."""
        self.hierarchy = {}

        def _add_hierarchy_level(node_dict, level, parent_idx):
            """Recursively add hierarchy levels."""
            lk = 'level-%d' % (level)

            child_nodes = [
                {'nidx': k, 'parent_idx': parent_idx,
                 'is_leaf': not isinstance(v, dict)} for k, v in node_dict.items()]

            if len(child_nodes) < 1:
                return

            # add to current level
            if lk in self.hierarchy.keys():
                self.hierarchy[lk].extend(child_nodes)
            else:
                self.hierarchy[lk] = child_nodes

            # add next level
            for nidx, ndict in node_dict.items():
                if isinstance(ndict, dict):
                    _add_hierarchy_level(ndict, level=level + 1, parent_idx=nidx)

        _add_hierarchy_level(self.tree_dict, level=0, parent_idx=0)

    def _merge_polygons(self, poly_list):
        if self.buffer_size > 0:
            poly_list = [j.buffer(self.buffer_size) for j in poly_list]
            merged_polys = unary_union(poly_list).buffer(-self.buffer_size)
        else:
            merged_polys = unary_union(poly_list)
        return merged_polys

    def _merge_leafs(self, leafs):
        nest_polygons = []
        for leaf in leafs:
            leafidx = int(leaf.split('polygon-')[1])
            nest = dict(self.contours_slice.loc[leafidx, :])
            coords = _parse_annot_coords(nest)
            nest_polygons.append(Polygon(coords))
        return self._merge_polygons(nest_polygons)

    def _get_merged_polygon(self, nidx):
        self.rtree.cursor._become(nidx)
        leafs = [c.leaf_obj() for c in self.rtree.cursor.children()]
        merged_polygon = self._merge_leafs(leafs)
        return merged_polygon

    def get_merged_multipolygon(self):
        """Get final merged shapely multipolygon by hierarchical merger."""
        merged_polygons_all = {}

        for level in range(len(self.hierarchy) - 1, -1, -1):

            merged_polygons = {}

            # merge polygons from previous level
            to_merge = {}
            for node in self.hierarchy['level-%d' % level]:
                if not node['is_leaf']:
                    if node['parent_idx'] not in to_merge.keys():
                        to_merge[node['parent_idx']] = []
                    to_merge[node['parent_idx']].append(merged_polygons_all[
                        'level-%d' % (level + 1)][node['nidx']])
                    del merged_polygons_all[
                        'level-%d' % (level + 1)][node['nidx']]

            for parent_idx, polygon_list in to_merge.items():
                merged_polygons[parent_idx] = self._merge_polygons(
                    polygon_list)

            # merge polygons from this level
            to_merge = {}
            for node in self.hierarchy['level-%d' % level]:
                if node['is_leaf']:
                    if node['parent_idx'] not in to_merge.keys():
                        to_merge[node['parent_idx']] = []
                    self.rtree.cursor._become(node['nidx'])
                    to_merge[node['parent_idx']].append(
                        self.rtree.cursor.leaf_obj())

            for parent_idx, leafs in to_merge.items():
                merged_polygons[parent_idx] = self._merge_leafs(leafs)

            # assign to persistent dict
            merged_polygons_all['level-%d' % level] = merged_polygons

        return merged_polygons_all['level-0'][0]

    def _get_coord_str_from_polygon(self, polygon):
        """Parse shapely polygon coordinates into string form (Internal)."""
        coords = np.int32(polygon.exterior.coords.xy)
        coords_x = ','.join([str(j) for j in coords[0, :]])
        coords_y = ','.join([str(j) for j in coords[1, :]])
        return coords_x, coords_y, coords.T

    def _add_single_merged_edge_contour(self, polygon, group):
        """Add single contour to self.new_contours (Internal)."""
        idx = self.new_contours.shape[0]
        self.new_contours.loc[idx, 'type'] = 'polyline'
        self.new_contours.loc[idx, 'group'] = group
        self.new_contours.loc[idx, 'has_holes'] = int(
            polygon.boundary.geom_type == 'MultiLineString')
        coords_x, coords_y, coords = self._get_coord_str_from_polygon(
            polygon)
        self.new_contours.loc[idx, 'coords_x'] = coords_x
        self.new_contours.loc[idx, 'coords_y'] = coords_y
        xmin, ymin = np.min(coords, axis=0)
        xmax, ymax = np.max(coords, axis=0)
        self.new_contours.loc[idx, 'xmin'] = xmin
        self.new_contours.loc[idx, 'ymin'] = ymin
        self.new_contours.loc[idx, 'xmax'] = xmax
        self.new_contours.loc[idx, 'ymax'] = ymax
        self.new_contours.loc[idx, 'bbox_area'] = int(
            (ymax - ymin) * (xmax - xmin))

    def _add_merged_multipolygon_contours(
            self, merged_multipolygon, group, monitorPrefix=''):
        """Add merged polygons to self.new_contours df (Internal)."""
        if merged_multipolygon.geom_type == 'Polygon':
            merged_multipolygon = [merged_multipolygon]
        elif (merged_multipolygon.geom_type == 'MultiPolygon' and
                hasattr(merged_multipolygon, 'geoms')):
            merged_multipolygon = merged_multipolygon.geoms
        for pno, polygon in enumerate(merged_multipolygon):
            self._print2('%s: contour %d of %d' % (
                monitorPrefix, pno + 1, len(merged_multipolygon)))
            self._add_single_merged_edge_contour(polygon, group=group)

    def run_for_single_group(self, group, monitorPrefix=''):
        """Run sequence for merging polygons & adding contours (one group)."""
        # Prep to get polygons
        self._print1('%s: set_contours_slice' % monitorPrefix)
        self.set_contours_slice(group)
        self._print1('%s: create_rtree' % monitorPrefix)
        self.create_rtree()
        self._print1('%s: set_tree_dict' % monitorPrefix)
        self.set_tree_dict()
        self._print1('%s: set_hierarchy' % monitorPrefix)
        self.set_hierarchy()

        # get shapely multipolygon object with merged adjacent contours
        self._print1('%s: get_merged_multipolygon' % monitorPrefix)
        merged_multipolygon = self.get_merged_multipolygon()

        # add contours to new dataframe
        self._print1('%s: _add_merged_multipolygon_contours' % monitorPrefix)
        self._add_merged_multipolygon_contours(
            merged_multipolygon, group=group)

    def run(self):
        """Run sequence for merging polygons & adding contours."""
        for group in self.unique_groups:
            monitorPrefix = '%s: %s' % (self.monitorPrefix, group)
            self.run_for_single_group(group, monitorPrefix=monitorPrefix)



"""
This package contains functions to enhance edges in images.
"""
# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .gaussian_grad import gaussian_grad

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'gaussian_grad',
)



import json
import logging
from pathlib import Path

import large_image
import numpy as np

import histomicstk
import histomicstk.preprocessing.color_deconvolution as htk_cd
from histomicstk.cli import utils
from histomicstk.cli.utils import CLIArgumentParser

logging.basicConfig()


def colorDeconvolve(args):
    sink0 = large_image.new()
    sink1 = large_image.new()
    sink2 = large_image.new()
    ts = large_image.getTileSource(args.inputImageFile)
    region = {}
    it_kwargs = {}
    tileSize = 8192
    it_kwargs['tile_size'] = dict(width=tileSize, height=tileSize)

    # Provides crop area if ROI present in arguments
    if np.all(np.array(args.region) != -1):
        it_kwargs['region'] = {
            'left': args.region[0],
            'top': args.region[1],
            'width': args.region[2],
            'height': args.region[3],
            'units': 'base_pixels',
        }
        sink0.crop = tuple(args.region)
        sink1.crop = tuple(args.region)
        sink2.crop = tuple(args.region)
        region = utils.get_region_dict(args.region, None, ts)['region']

    # Create stain matrix
    print('>> Creating stain matrix')

    w = utils.get_stain_matrix(args)
    print(w)

    # Perform color deconvolution
    print('>> Performing color deconvolution')
    for tile in ts.tileIterator(**it_kwargs):
        im_stains = htk_cd.color_deconvolution(tile['tile'], w).Stains
        sink0.addTile(im_stains[:, :, 0], tile['x'], tile['y'])
        sink1.addTile(im_stains[:, :, 1], tile['x'], tile['y'])
        sink2.addTile(im_stains[:, :, 2], tile['x'], tile['y'])

    # Write stain images to output
    print('>> Outputting individual stain images')

    print(args.outputStainImageFile_1)
    sink0.write(args.outputStainImageFile_1)
    print(args.outputStainImageFile_2)
    sink1.write(args.outputStainImageFile_2)
    print(args.outputStainImageFile_3)
    sink2.write(args.outputStainImageFile_3)

    return region


def main(args):

    # Read Input Image
    print('>> Reading input image')

    print(args.inputImageFile)

    region = colorDeconvolve(args)

    if args.outputAnnotationFile:

        annotation = [{
            'name': 'Deconvolution %s' % (
                args.stain_1 if args.stain_1 != 'custom' else str(args.stain_1_vector)),
            'elements': [{
                'type': 'image',
                'girderId': 'outputStainImageFile_1',
                'transform': {
                    'xoffset': region.get('left', 0),
                    'yoffset': region.get('top', 0),
                },
            }],
            'attributes': {
                'cli': Path(__file__).stem,
                'params': vars(args),
                'version': histomicstk.__version__,
            },
        }, {
            'name': 'Deconvolution %s' % (
                args.stain_2 if args.stain_2 != 'custom' else str(args.stain_2_vector)),
            'elements': [{
                'type': 'image',
                'girderId': 'outputStainImageFile_2',
                'transform': {
                    'xoffset': region.get('left', 0),
                    'yoffset': region.get('top', 0),
                },
            }],
            'attributes': {
                'cli': Path(__file__).stem,
                'params': vars(args),
                'version': histomicstk.__version__,
            },
        }, {
            'name': 'Deconvolution %s' % (
                args.stain_3 if args.stain_3 != 'custom' else str(args.stain_3_vector)),
            'elements': [{
                'type': 'image',
                'girderId': 'outputStainImageFile_3',
                'transform': {
                    'xoffset': region.get('left', 0),
                    'yoffset': region.get('top', 0),
                },
            }],
            'attributes': {
                'cli': Path(__file__).stem,
                'params': vars(args),
                'version': histomicstk.__version__,
            },
        }]
        if args.stain_3 == 'null':
            annotation[2:] = []
        if args.stain_2 == 'null':
            annotation[1:2] = []

        with open(args.outputAnnotationFile, 'w') as annotation_file:
            json.dump(annotation, annotation_file, separators=(',', ':'), sort_keys=False)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



import numpy as np


def perimeter(im_label, conn=4):
    """Converts a label or binary mask image to a binary perimeter image.

    Uses 4-neighbor or 8-neighbor shifts to detect pixels whose values do
    not agree with their neighbors.

    Parameters
    ----------
    im_label : array_like
        A label or binary mask image.
    conn : double or int
        Neighborhood connectivity to evaluate. Valid values are 4 or 8.
        Default value = 4.

    Returns
    -------
    Mask : array_like
        A binary image where object perimeter pixels have value 1, and
        non-perimeter pixels have value 0.

    See Also
    --------
    histomicstk.segmentation.embed_boundaries

    """
    # initialize temporary variable
    Mask = np.zeros(im_label.shape)
    Temp = np.zeros(im_label.shape)

    # check left-right neighbors
    Temp[:, 0:-2] = np.not_equal(im_label[:, 0:-2], im_label[:, 1:-1])
    Temp[:, 1:-1] = np.logical_or(Temp[:, 1:-1], Temp[:, 0:-2])
    Mask = np.logical_or(Mask, Temp)

    # check up-down neighbors
    Temp[0:-2, :] = np.not_equal(im_label[0:-2, :], im_label[1:-1, :])
    Temp[1:-1, :] = np.logical_or(Temp[1:-1, :], Temp[0:-2, :])
    Mask = np.logical_or(Mask, Temp)

    # additional calculations if conn == 8
    if conn == 8:

        # slope 1 diagonal shift
        Temp[1:-1, 0:-2] = np.not_equal(im_label[0:-2, 1:-2], im_label[1:-1, 0:-2])
        Temp[0:-2, 1:-1] = np.logical_or(Temp[0:-2, 1:-1], Temp[1:-1, 0:-2])
        Mask = np.logical_or(Mask, Temp)

        # slope -1 diagonal shift
        Temp[1:-1, 1:-1] = np.not_equal(im_label[0:-2, 0:-2], im_label[1:-1, 1:-1])
        Temp[0:-2, 0:-2] = np.logical_or(Temp[0:-2, 0:-2], Temp[1:-1, 1:-1])
        Mask = np.logical_or(Mask, Temp)

    # generate label-valued output
    return Mask.astype(np.uint32) * im_label



import collections

import numpy as np


def gaussian_grad(im_input, sigma):
    """Performs smoothing with derivative gaussian kernel.

    Uses separable convolution to simultaneously smooth and calculate the
    gradient of a grayscale image.

    Parameters
    ----------
    im_input : array_like
        An intensity image.
    sigma : double
        Standard deviation of smoothing kernel used in gradient calculation.

    Returns
    -------
    dx : array_like
        An intensity image of the X gradient component.
    dy : array_like
        An intensity image of the Y gradient component.

    Notes
    -----
    Return values are returned as a namedtuple

    """
    import scipy.signal as signal

    # generate separable gaussian derivative kernels
    x = np.linspace(0, np.ceil(2 * 3 * sigma), int(np.ceil(2 * 3 * sigma) + 1))
    y = np.linspace(0, np.ceil(2 * 3 * sigma), int(np.ceil(2 * 3 * sigma) + 1))
    x -= np.ceil(2 * 3 * sigma) / 2  # center independent variables at zero
    y -= np.ceil(2 * 3 * sigma) / 2
    x = np.reshape(x, (1, x.size))  # reshape to 2D row and column vectors
    y = np.reshape(y, (y.size, 1))
    xGx = 2 * x / (sigma ** 2) * np.exp(-x ** 2 / (2 * sigma ** 2)) \
        / (sigma * (2 * np.pi) ** 0.5)
    yGx = np.exp(-y**2 / (2 * sigma ** 2)) / ((2 * np.pi) ** 0.5 * sigma)
    xGy = np.exp(-x**2 / (2 * sigma ** 2)) / ((2 * np.pi) ** 0.5 * sigma)
    yGy = 2 * y / (sigma ** 2) * np.exp(-y ** 2 / (2 * sigma ** 2)) \
        / (sigma * (2 * np.pi) ** 0.5)

    # smoothed gradients of input image
    dx = signal.convolve2d(im_input, xGx, mode='same')
    dx = signal.convolve2d(dx, yGx, mode='same')
    dy = signal.convolve2d(im_input, xGy, mode='same')
    dy = signal.convolve2d(dy, yGy, mode='same')

    # format output
    Output = collections.namedtuple('Output', ['dX', 'dY'])
    Gradients = Output(dx, dy)

    return Gradients



def convert_matrix_to_image(m, shape):
    """Convert a column matrix of pixels to a 3D image given by shape.
    The number of channels is taken from m, not shape.  If shape has
    length 2, the matrix is returned unchanged.  This is the inverse
    of convert_image_to_matrix:

    im == convert_matrix_to_image(convert_image_to_matrix(im),
    im.shape)

    """
    if len(shape) == 2:
        return m

    return m.T.reshape(shape[:-1] + (m.shape[0],))



import numpy as np

from ._compute_marginal_glcm_probs_cython import \
    _compute_marginal_glcm_probs_cython
from .graycomatrixext import (_default_num_levels, _default_offsets,
                              graycomatrixext)


def compute_haralick_features(im_label, im_intensity, offsets=None,
                              num_levels=None, gray_limits=None, rprops=None):
    r"""
    Calculates 26 Haralick texture features for each object in the given label
    mask.

    These features are derived from gray-level co-occurence matrix (GLCM)
    that is a two dimensional histogram containing the counts/probabilities of
    co-occurring intensity values with a given neighborhood offset in the
    region occupied by an object in the image.

    Parameters
    ----------
    im_label : array_like
        An ND labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.

    im_intensity : array_like
        An ND single channel intensity image.

    offsets : array_like, optional
        A (num_offsets, num_image_dims) array of offset vectors
        specifying the distance between the pixel-of-interest and
        its neighbor. Note that the first dimension corresponds to
        the rows.

        See `histomicstk.features.graycomatrixext` for more details.

    num_levels : unsigned int, optional
        An integer specifying the number of gray levels For example, if
        `NumLevels` is 8,  the intensity values of the input image are
        scaled so they are integers between 1 and 8.  The number of gray
        levels determines the size of the gray-level co-occurrence matrix.

        Default: 2 for binary/logical image, 32 for numeric image

    gray_limits : array_like, optional
        A two-element array specifying the desired input intensity range.
        Intensity values in the input image will be clipped into this range.

        Default: [0, 1] for boolean-valued image, [0, 255] for integer-valued
        image, and [0.0, 1.0] for-real valued image

    Returns
    -------
    fdata: pandas.DataFrame
        A pandas dataframe containing the haralick features.

    Notes
    -----
    This function computes the following list of haralick features derived
    from normalized GLCMs (P) of the given list of neighborhood offsets:

    Haralick.ASM.Mean, Haralick.ASM.Range : float
        Mean and range of the angular second moment (ASM) feature for GLCMs
        of all offsets. It is a measure of image homogeneity and is computed
        as follows:

        .. math::

            ASM = \sum_{i,j=0}^{levels-1} p(i,j)^2

    Haralick.Contrast.Mean, Haralick.Contrast.Range : float
        Mean and range of the Contrast feature for GLCMs of all offsets. It is
        a measure of the amount of variation between intensities of
        neighboiring pixels. It is equal to zero for a constant image and
        increases as the amount of variation increases. It is computed as
        follows:

        .. math::

            Contrast = \sum_{i,j=0}^{levels-1}  (i-j)^2 p(i,j)

    Haralick.Correlation.Mean, Haralick.Correlation.Range : float
        Mean and range of the Correlation feature for GLCMs of all offsets. It
        is a measure of correlation between the intensity values of
        neighboring pixels. It is computed as follows:

        .. math::

            Correlation =
            \sum_{i,j=0}^{levels-1} p(i,j)\left[\frac{(i-\mu_i)
            (j-\mu_j)}{\sigma_i \sigma_j}\right]

    Haralick.SumOfSquares.Mean, Haralick.SumOfSquares.Range : float
        Mean and range of the SumOfSquares feature for GLCMs of all offsets.
        It is a measure of variance and is computed as follows:

        .. math::

            SumofSquare =
            \sum_{i,j=0}^{levels-1} (i - \mu)^2 p(i,j)

    Haralick.IDM.Mean, Haralick.IDM.Range : float
        Mean and range of the inverse difference moment (IDM) feature for
        GLCMS of all offsets. It is a measure of homogeneity and is computed
        as follows:

        .. math::
            IDM = \sum_{i,j=0}^{levels-1} \frac{1}{1 + (i - j)^2} p(i,j)

    Haralick.SumAverage.Mean, Haralick.SumAverage.Range : float
        Mean and range of sum average feature for GLCMs of all offsets.
        It is computed as follows:

        .. math::

            SumAverage = \sum_{k=2}^{2 levels} k p_{x+y}(k), \qquad where \\

            p_{x+y}(k) =
            \sum_{i,j=0}^{levels-1} \delta_{i+j, k} p(i,j) \\

            \delta_{m,n} = \left\{
            \begin{array}{11}
                1 & {\rm when ~} m=n \\
                0 & {\rm when ~} m \ne n
            \end{array}
            \right.

    Haralick.SumVariance.Mean, Haralick.SumVariance.Range : float
        Mean and range of sum variance feature for the GLCMS of all offsets.
        It is computed as follows:

        .. math::

            SumVariance =
            \sum_{k=2}^{2 levels} (k - SumEntropy) p_{x+y}(k)

    Haralick.SumEntropy.Mean, Haralick.SumEntropy.Range : float
        Mean and range of the sum entropy features for GLCMS of all offsets.
        It is computed as follows:

        .. math::

            SumEntropy =
            - \sum_{k=2}^{2 levels} p_{x+y}(k) \log(p_{x+y}(k))

    Haralick.Entropy.Mean, Haralick.Entropy.Range : float
        Mean and range of the entropy features for GLCMs of all offsets.
        It is computed as follows:

        .. math::

            Entropy = - \sum_{i,j=0}^{levels-1} p(i,j) \log(p(i,j))


    Haralick.DifferenceVariance.Mean, Haralick.DifferenceVariance.Range : float
        Mean and Range of the difference variance feature of GLCMs of all
        offsets. It is computed as follows:

        .. math::

            DifferenceVariance = {\rm variance \ of ~} p_{x-y}, \qquad where \\

            p_{x-y}(k) =
            \sum_{i,j=0}^{levels-1} \delta_{|i-j|, k} p(i,j)

    Haralick.DifferenceEntropy.Mean, Haralick.DifferenceEntropy.Range : float
        Mean and range of the difference entropy feature for GLCMS of all
        offsets. It is computed as follows:

        .. math::

            DifferenceEntropy = {\rm entropy \ of ~} p_{x-y}

    Haralick.IMC1.Mean, Haralick.IMC1.Range : float
        Mean and range of the first information measure of correlation
        feature for GLCMs of all offsets. It is computed as follows:

        .. math::

            IMC1 = \frac{HXY - HXY1}{\max(HX,HY)}, \qquad where \\

            HXY = -\sum_{i,j=0}^{levels-1} p(i,j) \log(p(i,j)) \\

            HXY1 = -\sum_{i,j=0}^{levels-1} p(i,j) \log(p_x(i) p_y(j)) \\

            HX = -\sum_{i=0}^{levels-1} p_x(i) \log(p_x(i))    \\

            HY = -\sum_{j=0}^{levels-1} p_y(j) \log(p_y(j))    \\

            p_x(i) = \sum_{j=1}^{levels} p(i,j) \\

            p_y(j) = \sum_{j=1}^{levels} p(i,j)

    Haralick.IMC2.Mean, Haralick.IMC2.Range : float
        Mean and range of the second information measure of correlation
        feature for GLCMs of all offsets. It is computed as follows:

        .. math::

            IMC2 = [1 - \exp(-2(HXY2 - HXY))]^{1/2}, \qquad where \\

            HXY2 = -\sum_{i,j=0}^{levels-1} p_x(i) p_y(j) \log(p_x(i) p_y(j))

    References
    ----------
    .. [#] Haralick, et al. "Textural features for image classification,"
       IEEE Transactions on Systems, Man, and Cybernatics, vol. 6,
       pp: 610-621, 1973.
    .. [#] Luis Pedro Coelho. "Mahotas: Open source software for scriptable
       computer vision," Journal of Open Research Software, vol 1, 2013.

    """
    import pandas as pd
    from skimage.measure import regionprops

    # List of feature names
    feature_list = [
        'Haralick.ASM',
        'Haralick.Contrast',
        'Haralick.Correlation',
        'Haralick.SumOfSquares',
        'Haralick.IDM',
        'Haralick.SumAverage',
        'Haralick.SumVariance',
        'Haralick.SumEntropy',
        'Haralick.Entropy',
        'Haralick.DifferenceVariance',
        'Haralick.DifferenceEntropy',
        'Haralick.IMC1',
        'Haralick.IMC2',
    ]

    agg_feature_list = []
    for fname in feature_list:
        agg_feature_list.append(fname + '.Mean')
        agg_feature_list.append(fname + '.Range')

    # num_levels
    if num_levels is None:
        num_levels = _default_num_levels(im_intensity)

    # check for consistent shapes between 'I' and 'Label'
    if im_intensity.shape != im_label.shape:
        msg = "Inputs 'I' and 'Label' must have same shape"
        raise ValueError(msg)

    num_dims = len(im_intensity.shape)

    # offsets
    if offsets is None:

        # set default offset value
        offsets = _default_offsets(im_intensity)

    else:

        # check sanity
        if offsets.shape[1] != num_dims:
            msg = 'Dimension mismatch between input image and offsets'
            raise ValueError(
                msg,
            )

    num_offsets = offsets.shape[0]

    # compute object properties if not provided
    if rprops is None:
        rprops = regionprops(im_label)

    # create pandas data frame containing the features for each object
    numLabels = len(rprops)
    fdata = pd.DataFrame(np.zeros((numLabels, len(agg_feature_list))),
                         columns=agg_feature_list)

    n_Minus = np.arange(num_levels)
    n_Plus = np.arange(2 * num_levels - 1)

    x, y = np.mgrid[0:num_levels, 0:num_levels]
    xy = x * y
    xy_IDM = 1. / (1 + np.square(x - y))

    e = 0.00001  # small positive constant to avoid log 0
    eps = np.finfo(float).eps  # small constant to avoid div / 0

    for i in range(numLabels):
        if rprops[i] is None:
            continue

        # get bounds of an intensity image
        minr, minc, maxr, maxc = rprops[i].bbox

        # grab nucleus mask
        subImage = im_intensity[minr:maxr + 1, minc:maxc + 1]

        # gets GLCM or gray-tone spatial dependence matrix
        arrayGLCM = graycomatrixext(subImage, offsets=offsets,
                                    num_levels=num_levels,
                                    gray_limits=gray_limits,
                                    symmetric=True, normed=True)

        # Compute haralick features for each offset
        ldata = pd.DataFrame(np.zeros((num_offsets, len(feature_list))),
                             columns=feature_list)

        for r in range(num_offsets):

            nGLCM = arrayGLCM[:, :, r]

            # get marginal-probabilities
            px, py, pxPlusy, pxMinusy = _compute_marginal_glcm_probs_cython(
                nGLCM)

            # computes angular second moment
            ldata.at[r, 'Haralick.ASM'] = np.sum(np.square(nGLCM))

            # computes contrast
            ldata.at[r, 'Haralick.Contrast'] = \
                np.dot(np.square(n_Minus), pxMinusy)

            # computes correlation
            # gets weighted mean and standard deviation of px and py
            meanx = np.dot(n_Minus, px)
            variance = np.dot(px, np.square(n_Minus)) - np.square(meanx)
            nGLCMr = np.ravel(nGLCM)

            har_corr = (np.dot(np.ravel(xy), nGLCMr) - np.square(meanx)) /\
                max(eps, variance)
            ldata.at[r, 'Haralick.Correlation'] = np.clip(har_corr,
                                                          a_min=-1, a_max=1)

            # computes sum of squares : variance
            ldata.at[r, 'Haralick.SumOfSquares'] = variance

            # computes inverse difference moment
            ldata.at[r, 'Haralick.IDM'] = \
                np.dot(np.ravel(xy_IDM), nGLCMr)

            # computes sum average
            ldata.at[r, 'Haralick.SumAverage'] = \
                np.dot(n_Plus, pxPlusy)

            # computes sum variance
            # [1] uses sum entropy, but we use sum average
            ldata.at[r, 'Haralick.SumVariance'] = \
                np.dot(np.square(n_Plus), pxPlusy) - \
                np.square(ldata.at[r, 'Haralick.SumAverage'])

            # computes sum entropy
            ldata.at[r, 'Haralick.SumEntropy'] = \
                -np.dot(pxPlusy, np.log2(pxPlusy + e))

            # computes entropy
            ldata.at[r, 'Haralick.Entropy'] = \
                -np.dot(nGLCMr, np.log2(nGLCMr + e))

            # computes variance px-y
            ldata.at[r, 'Haralick.DifferenceVariance'] = np.var(pxMinusy)

            # computes difference entropy px-y
            ldata.at[r, 'Haralick.DifferenceEntropy'] = \
                -np.dot(pxMinusy, np.log2(pxMinusy + e))

            # computes information measures of correlation
            # gets entropies of px and py
            HX = -np.dot(px, np.log2(px + e))
            HY = -np.dot(py, np.log2(py + e))
            HXY = ldata.at[r, 'Haralick.Entropy']
            pxy_ij = np.outer(px, py)
            pxy_ijr = np.ravel(pxy_ij)
            HXY1 = -np.dot(nGLCMr, np.log2(pxy_ijr + e))
            HXY2 = -np.dot(pxy_ijr, np.log2(pxy_ijr + e))
            ldata.at[r, 'Haralick.IMC1'] = ((HXY - HXY1) / max(HX, HY)) if max(HX, HY) else 0

            # computes information measures of correlation
            ldata.at[r, 'Haralick.IMC2'] = \
                np.sqrt(max(0, 1 - np.exp(-2.0 * (HXY2 - HXY))))

        fdata.values[i, ::2] = np.mean(ldata.values, axis=0)
        fdata.values[i, 1::2] = np.ptp(ldata.values, axis=0)

    return fdata



import copy
import os
import shutil
import sys
import tempfile

import numpy as np
import pytest
from imageio import imread
from pandas import read_csv

from histomicstk.annotations_and_masks.annotation_and_mask_utils import (
    get_bboxes_from_slide_annotations, get_scale_factor_and_appendStr,
    scale_slide_annotations)
from histomicstk.annotations_and_masks.annotations_to_object_mask_handler import (
    annotations_to_contours_no_mask, get_all_rois_from_slide_v2)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
import htk_test_utilities as utilities  # noqa
from htk_test_utilities import getTestFilePath, girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.slide_annotations = None
        self.element_infos = None
        self.BASE_SAVEPATH = None
        self.SAVEPATHS = None
        self.test_annots_to_contours_kwargs = None
        self.get_all_rois_kwargs = None
        self.MPP = None
        self.MAG = None


cfg = Cfg()

# pytest runs tests in the order they appear in the module


@pytest.mark.usefixtures('girderClient')  # noqa
def test_prep(girderClient):  # noqa

    cfg.gc = girderClient

    iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]

    # read GTCodes dataframe
    gtcodePath = getTestFilePath('sample_GTcodes.csv')
    GTCodes_dict = read_csv(gtcodePath)
    GTCodes_dict.index = GTCodes_dict.loc[:, 'group']
    GTCodes_dict = GTCodes_dict.to_dict(orient='index')

    # just a temp directory to save masks for now
    cfg.BASE_SAVEPATH = tempfile.mkdtemp()
    cfg.SAVEPATHS = {
        'contours': os.path.join(cfg.BASE_SAVEPATH, 'contours'),
        'rgb': os.path.join(cfg.BASE_SAVEPATH, 'rgbs'),
        'visualization': os.path.join(cfg.BASE_SAVEPATH, 'vis'),
        'mask': os.path.join(cfg.BASE_SAVEPATH, 'masks'),
    }
    for _, savepath in cfg.SAVEPATHS.items():
        if not os.path.exists(savepath):
            os.mkdir(savepath)

    # Microns-per-pixel / Magnification (either or)
    cfg.MPP = 5.0
    cfg.MAG = None

    # get annotations for slide
    cfg.slide_annotations = cfg.gc.get('/annotation/item/' + iteminfo['_id'])

    # scale up/down annotations by a factor
    sf, _ = get_scale_factor_and_appendStr(
        gc=cfg.gc, slide_id=iteminfo['_id'], MPP=cfg.MPP, MAG=cfg.MAG)
    cfg.slide_annotations = scale_slide_annotations(cfg.slide_annotations, sf=sf)

    # get bounding box information for all annotations
    cfg.element_infos = get_bboxes_from_slide_annotations(cfg.slide_annotations)

    # common params for annotations_to_contours_no_mask()
    annotations_to_contours_kwargs = {
        'MPP': cfg.MPP, 'MAG': cfg.MAG,
        'linewidth': 0.2,
        'get_rgb': True, 'get_visualization': True,
    }

    # params for TESTING annotations_to_contours_no_mask()
    cfg.test_annots_to_contours_kwargs = copy.deepcopy(
        annotations_to_contours_kwargs)
    cfg.test_annots_to_contours_kwargs.update({
        'gc': cfg.gc,
        'slide_id': iteminfo['_id'],
        'bounds': {
            'XMIN': 58000, 'XMAX': 63000,
            'YMIN': 35000, 'YMAX': 39000},
    })

    # params for getting all rois for slide
    cfg.get_all_rois_kwargs = {
        'gc': cfg.gc,
        'slide_id': iteminfo['_id'],
        'GTCodes_dict': GTCodes_dict,
        'save_directories': cfg.SAVEPATHS,
        'annotations_to_contours_kwargs': annotations_to_contours_kwargs,
        'slide_name': 'TCGA-A2-A0YE',
        'verbose': False,
        'monitorprefix': 'test',
    }


class TestGetSlideRegionNoMask:
    """Test methods for getting ROI contours from annotations."""

    def test_annotations_to_contours_no_mask_1(self):
        """Test annotations_to_contours_no_mask()."""
        # get specified region -- without providing scaled annotations
        roi_out_1 = annotations_to_contours_no_mask(
            mode='manual_bounds', **cfg.test_annots_to_contours_kwargs)

        # get specified region -- with providing scaled annotations
        roi_out_2 = annotations_to_contours_no_mask(
            mode='manual_bounds', slide_annotations=cfg.slide_annotations,
            element_infos=cfg.element_infos, **cfg.test_annots_to_contours_kwargs)

        for roi_out in (roi_out_1, roi_out_2):
            assert set(roi_out.keys()) == {
                'bounds', 'rgb', 'contours', 'visualization'}
            assert roi_out['rgb'].shape == (200, 251, 3)
            assert roi_out['visualization'].shape == (200, 251, 3)
            assert len(roi_out['contours']) > 56
            assert (
                len(roi_out['contours']) < 68)
            assert set(roi_out['contours'][0].keys()) == {
                'annidx', 'elementidx', 'element_girder_id', 'type', 'label',
                'annotation_girder_id', 'bbox_area', 'group', 'color',
                'ymin', 'ymax', 'xmin', 'xmax', 'coords_x', 'coords_y',
            }

    def test_annotations_to_contours_no_mask_2(self):
        """Test get_image_and_mask_from_slide()."""
        # get ROI bounding everything
        roi_out = annotations_to_contours_no_mask(
            mode='min_bounding_box', slide_annotations=cfg.slide_annotations,
            element_infos=cfg.element_infos, **cfg.test_annots_to_contours_kwargs)

        assert set(roi_out.keys()) == {
            'bounds', 'rgb', 'contours', 'visualization'}
        assert roi_out['rgb'].shape == (321, 351, 3)
        assert roi_out['visualization'].shape == (321, 351, 3)
        assert len(roi_out['contours']) > 72
        assert (
            len(roi_out['contours']) < 80)
        assert set(roi_out['contours'][0].keys()) == {
            'annidx', 'elementidx', 'element_girder_id', 'type', 'label',
            'annotation_girder_id', 'bbox_area', 'group', 'color',
            'ymin', 'ymax', 'xmin', 'xmax', 'coords_x', 'coords_y',
        }

    def test_get_all_rois_from_slide_v2(self):
        """Test get_all_rois_from_slide_v2()."""
        # First we test the object segmentation mode
        cfg.get_all_rois_kwargs['mode'] = 'object'
        savenames = get_all_rois_from_slide_v2(**cfg.get_all_rois_kwargs)

        # basic checks
        assert len(savenames) == 3
        assert set(savenames[0].keys()) == {
            'mask', 'rgb', 'visualization', 'contours'}
        assert {
            'TCGA-A2-A0YE_left-58463_top-38203_bottom-39760_right-60379.png',
            'TCGA-A2-A0YE_left-59181_top-33473_bottom-38043_right-63712.png',
            'TCGA-A2-A0YE_left-57584_top-35788_bottom-37425_right-59421.png',
        }, {os.path.basename(savename['mask']) for savename in savenames}

        # shape & value check
        imname = 'TCGA-A2-A0YE_left-57584_top-35788_bottom-37425_right-59421'
        mask = imread(os.path.join(cfg.SAVEPATHS['mask'], imname + '.png'))
        assert mask.shape == (82, 92, 3)
        assert set(np.unique(mask[..., 0])) == {0, 1, 2, 7}
        assert set(np.unique(mask[..., 1])) == {0, 1}
        assert set(np.unique(mask[..., 2])) == {
            0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12} or set(np.unique(mask[..., 2])) == {
            0, 2, 3, 4, 5, 6, 7, 8}

        # Second, we test the semantic segmentation mode
        cfg.get_all_rois_kwargs['mode'] = 'semantic'
        savenames = get_all_rois_from_slide_v2(**cfg.get_all_rois_kwargs)
        assert len(savenames) == 3

        # shape check
        mask = imread(os.path.join(cfg.SAVEPATHS['mask'], imname + '.png'))
        assert mask.shape == (82, 92)

        # cleanup
        shutil.rmtree(cfg.BASE_SAVEPATH)



import numpy as np

from histomicstk.segmentation import label


def min_model(I, Delta=0.3, MaxLength=255, Compaction=3,
              MinArea=100, MinWidth=5, MinDepth=2, MinConcavity=np.inf):
    """Performs a nuclear segmentation using a gradient contour tracing and
    geometry splitting algorithm. Implemented from the reference below.

    Parameters
    ----------
    I : array_like
        An intensity image used for analyzing local minima/maxima and
        gradients. Dimensions M x N.
    Delta : float
        Fractional difference threshold between minima/maxima pairs to
        be included in seed point detection. Fractional difference
        ([0, 1]) in total image range e.g. Delta = 0.3 with a uint8
        input would translate to 0.3 * 255. Default value = 0.3.
    MaxLength : int
        Maximum allowable contour length. Default value = 255.
    Compaction : int
        Factor used in compacting objects to remove thin spurs. Referred to as
        'd' in the paper. Default value = 3.
    MinArea : int
        Minimum area of objects to analyze. Default value = 100.
    MinWidth : int
        Minimum max-width of objects to analyze. Default value = 5.
    MinDepth : float
        Minimum depth of concavities to consider during geometric splitting.
        Default value = 2.
    MinConcavity : float
        Minimum concavity score to consider when performing for geometric
        splitting. Default value = np.inf.

    Notes
    -----
    Objects are assumed to be dark (as nuclei in hematoxylin channel from color
    deconvolution). Smoothing improves accuracy and computation time by
    eliminating spurious seed points. Specifying a value for 'Delta' prevents
    shallow transitions from being included, also reducing computation time and
    increasing specificity.

    Returns
    -------
    X : array_like
        A 1D array of horizontal coordinates of contour seed pixels for
        tracing.
    Y : array_like
        A 1D array of the vertical coordinates of seed pixels for tracing.
    Min : array_like
        A 1D array of the corresponding minimum values for contour tracing of
        seed point X, Y.
    Max : array_like
        A 1D array of the corresponding maximum values for contour tracing of
        seed point X, Y.

    See Also
    --------
    histomicstk.segmentation.label.trace_object_boundaries

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
       Virtual Microscopy Images: A Minimum-Model Approach" in Nature
       Scientific Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    # identify contour seed points
    X, Y, Min, Max = seed_contours(I, Delta)

    # trace contours from seeds
    cXs, cYs = trace_contours(I, X, Y, Min, Max, MaxLength=255)

    # score successfully traced contours
    Scores = score_contours(I, cXs, cYs)

    # construct label image from scored contours
    Label = label_contour(I.shape, cXs, cYs, Scores)

    # compact contours to remove spurs - the paper calls this "optimization"
    Label = label.compact(Label, Compaction)

    # cleanup label image
    Label = label.split(Label)
    Label = label.area_open(Label, MinArea)
    Label = label.width_open(Label, MinWidth)

    # split objects with concavities
    Label = split_concavities(Label, MinDepth, MinConcavity)

    return Label


def seed_contours(I, Delta=0.3):  # noqa
    """Detects seed pixels for contour tracing by finding max-gradient points
    between local minima and maxima in an intensity image.

    Parameters
    ----------
    I : array_like
        An intensity image used for analyzing local minima/maxima and
        gradients. Dimensions M x N.
    Delta : float
        Fractional difference threshold between minima/maxima pairs to
        be included in seed point detection. Fractional difference
        ([0, 1]) in total image range e.g. Delta = 0.3 with a uint8
        input would translate to 0.3 * 255.  Default value = 0.3.

    Notes
    -----
    Objects are assumed to be dark (as nuclei in hematoxylin channel from color
    deconvolution). Smoothing improves accuracy and computation time by
    eliminating spurious seed points. Specifying a value for 'Delta' prevents
    shallow transitions from being included, also reducing computation time and
    increasing specificity.

    Returns
    -------
    X : array_like
        A 1D array of horizontal coordinates of contour seed pixels for
        tracing.
    Y : array_like
        A 1D array of the vertical coordinates of seed pixels for tracing.
    Min : array_like
        A 1D array of the corresponding minimum values for contour tracing of
        seed point X, Y.
    Max : array_like
        A 1D array of the corresponding maximum values for contour tracing of
        seed point X, Y.

    See Also
    --------
    TraceBounds, SeedContours, MinimumModel

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    # initialize outputs
    X = []
    Y = []
    Min = []
    Max = []

    # for each image row
    for i in np.arange(I.shape[0]):

        # calculate gradient
        Gradient = np.hstack((np.nan, I[i, 2:] - I[i, 0:-2], np.nan))

        # identify local maxima and minimia of row 'i' of 'I'
        Maxima = ((I[i, 1:-1] >= I[i, 0:-2]) &
                  (I[i, 1:-1] > I[i, 2:])).nonzero()[0] + 1
        Minima = ((I[i, 1:-1] < I[i, 0:-2]) &
                  (I[i, 1:-1] <= I[i, 2:])).nonzero()[0] + 1

        # identify transitions - start of intervals of monotonic non-increase
        dI = np.sign(I[i, 1:].astype(float) - I[i, 0:-1].astype(float))
        dI = np.hstack((dI, dI[-1]))
        Transitions = np.nonzero(dI == 1)[0]
        Transitions = np.hstack((Transitions, I.shape[1] - 1))

        # define min/max neighbor pairs
        MinPair = []
        MaxPair = []
        if (Minima.size > 0) and (Maxima.size > 0):

            # initialize initial positions of min/max & transition indices
            MinPos = 0
            MaxPos = 0
            TranPos = 0

            # iterate through maxima, identifying relevant minima for each
            while MaxPos < Maxima.size:
                Index = np.nonzero(Minima > Maxima[MaxPos])[0]
                if Index.size:  # minima found beyond current maxima

                    # get position of next minimum in array 'Minima'
                    MinPos = Index[0]

                    # increment transition point to beyond current maxima
                    while (TranPos < Transitions.size) & \
                            (Transitions[TranPos] <= Maxima[MaxPos]):
                        TranPos += 1

                    # add minima to current maxima until transition is reached
                    while Minima[MinPos] <= Transitions[TranPos]:
                        MinPair.append(Minima[MinPos])
                        MaxPair.append(Maxima[MaxPos])
                        MinPos += 1
                        if MinPos == Minima.size:
                            break

                    # increment maxima
                    MaxPos += 1

                else:  # no minima beyond current maxima - quit
                    break

            # convert maxima/minima pairs to numpy arrays
            Maxima = np.asarray(MaxPair)
            Minima = np.asarray(MinPair)

            # remove pairs that do not have at least one pixel between them
            Close = ((Minima - Maxima) < 2).nonzero()
            Maxima = np.delete(Maxima, Close)
            Minima = np.delete(Minima, Close)

            # skip to next row if no min/max pairs exist after location filter
            if (Minima.size == 0) | (Maxima.size == 0):
                continue

            # remove pairs that do not have sufficient intensity transitions
            if Delta is not None:
                if np.issubdtype(I.dtype, np.integer):
                    Range = Delta * 255.0
                elif np.issubdtype(I.dtype, float):
                    Range = Delta * 1.0
                Shallow = (I[i, Maxima] - I[i, Minima] < Range).nonzero()
                Maxima = np.delete(Maxima, Shallow)
                Minima = np.delete(Minima, Shallow)

            # skip to next row if no min/max pairs exist after intensity filter
            if (Minima.size == 0) | (Maxima.size == 0):
                continue

            # identify max gradient locations within paired maxima/minima
            MinGrad = np.zeros(Maxima.shape, dtype=int)
            for j in np.arange(Maxima.size):
                MinGrad[j] = np.argmin(Gradient[Maxima[j] + 1:Minima[j]]) + \
                    Maxima[j] + 1

            # capture min, max values and add to list with seed coordinates
            if Maxima.size > 0:
                X.extend(MinGrad)
                Y.extend(i * np.ones(Maxima.size))
                Min.extend(I[i, Minima])
                Max.extend(I[i, MinGrad])

    # convert outputs from lists to numpy arrays
    X = np.array(X, dtype=np.uint)
    Y = np.array(Y, dtype=np.uint)
    Min = np.array(Min, dtype=I.dtype)
    Max = np.array(Max, dtype=I.dtype)

    # return seed pixels positions and intensity range intervals
    return X, Y, Min, Max


def trace_contours(I, X, Y, Min, Max, MaxLength=255):
    """Performs contour tracing of seed pixels in an intensity image using
    gradient information.

    Parameters
    ----------
    I : array_like
        An intensity image used for analyzing local minima/maxima and
        gradients. Dimensions M x N.
    X : array_like
        A 1D array of horizontal coordinates of contour seed pixels for
        tracing.
    Y : array_like
        A 1D array of the vertical coordinates of seed pixels for tracing.
    Min : array_like
        A 1D array of the corresponding minimum values for contour tracing of
        seed point X, Y.
    Max : array_like
        A 1D array of the corresponding maximum values for contour tracing of
        seed point X, Y.
    MaxLength : int
        Maximum allowable contour length. Default value = 255.

    Notes
    -----
    Can be computationally expensive for large numbers of contours. Use
    smoothing and delta thresholding when seeding contours to reduce burden.

    Returns
    -------
    cXs : list
        A list of 1D numpy arrays defining the horizontal coordinates of object
        boundaries.
    cYs : list
        A list of 1D numpy arrays defining the vertical coordinates of object
        boundaries.

    See Also
    --------
    SeedContours, ScoreContours, MinimumModel

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    # initialize list of lists containing contours
    cXs = []
    cYs = []

    # process each seed pixel sequentially
    for i in np.arange(X.size):

        # capture window surrounding (X[i], Y[i])
        W = I[max(0, Y[i] - np.ceil(MaxLength / 2.0)):
              min(I.shape[0] + 1, Y[i] + np.ceil(MaxLength / 2.0) + 1),
              max(0, X[i] - np.ceil(MaxLength / 2.0)):
              min(I.shape[1] + 1, X[i] + np.ceil(MaxLength / 2.0)) + 1]

        # binary threshold corresponding to seed pixel 'i'
        W = (W <= Max[i]) & (W >= Min[i])

        # embed with center pixel in middle of padded window
        Embed = np.zeros((W.shape[0] + 2, W.shape[1] + 2), dtype=bool)
        Embed[1:-1, 1:-1] = W

        # calculate location of (X[i], Y[i]) in 'Embed'
        pX = X[i] - max(0, X[i] - np.ceil(MaxLength / 2.0)) + 1
        pY = Y[i] - max(0, Y[i] - np.ceil(MaxLength / 2.0)) + 1

        # trace boundary, check stopping condition, append to list of contours
        cX, cY = label.trace_object_boundaries(Embed, conn=4,
                                               x_start=pX, y_start=pY,
                                               MaxLength=MaxLength)
        if cX[0][0] == cX[0][-1] and cY[0][0] == cY[0][-1] and\
                len(cX[0]) <= MaxLength:

            # add window offset to contour coordinates
            cX[0] = [
                x + max(0, X[i] - np.ceil(MaxLength / 2.0)) - 1 for x in cX[0]
            ]
            cY[0] = [
                y + max(0, Y[i] - np.ceil(MaxLength / 2.0)) - 1 for y in cY[0]
            ]

            # append to list of candidate contours
            cXs.append(np.array(cX[0], dtype=np.uint32))
            cYs.append(np.array(cY[0], dtype=np.uint32))

    return cXs, cYs


def score_contours(I, cXs, cYs):
    """Scores boundary contours using gradient information. Implemented from
    the reference below. Each contour is weighted by the average gradient and
    number of local gradient maxima along its path.

    Parameters
    ----------
    I : array_like
        An intensity image used for analyzing local minima/maxima and
        gradients. Dimensions M x N.
    cXs : list
        A list of 1D numpy arrays defining the horizontal coordinates of object
        boundaries.
    cYs : list
        A list of 1D numpy arrays defining the vertical coordinates of object
        boundaries.

    Notes
    -----
    Implemented from the reference below.

    Returns
    -------
    Scores : array_like
        A 1D array of horizontal coordinates of contour seed pixels for
        tracing.

    See Also
    --------
    TraceContours, LabelContour, MinimumModel

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    import scipy.ndimage as ndi

    # initialize output
    Scores = np.zeros(len(cXs))

    # generate Sobel filter response from input intensity image 'I'
    Gradients = ndi.sobel(I, mode='mirror')

    # generate local max in 3 x 3 window of Gradients
    Maxima = ndi.maximum_filter(Gradients, size=3, mode='mirror')

    # generate score for each contour
    for i in np.arange(len(cXs)):

        # get gradient pixels, local max gradient pixels
        cG = Gradients[cYs[i], cXs[i]]
        cMax = Maxima[cYs[i], cXs[i]]

        # calculate mean gradient
        MG = np.sum(np.abs(Gradients[cYs[i], cXs[i]])) / len(cXs[i])

        # calculate gradient fit
        GF = np.sum(cG == cMax) / len(cXs[i])

        # compute score as product of mean gradient and gradient fit
        Scores[i] = MG * GF

    return Scores


def label_contour(Shape, cXs, cYs, Scores):
    """Constructs a label image from scored contours. Masks for contours with
    low priority/score are placed first into the label image and then are
    overwritten by higher priority contours.

    Parameters
    ----------
    Shape : tuple
        The shape tuple of the desired label image (height, width).
    cXs : list
        A list of 1D numpy arrays defining the horizontal coordinates of object
        boundaries.
    cYs : list
        A list of 1D numpy arrays defining the vertical coordinates of object
        boundaries.
    Scores : array_like
        A 1D array of horizontal coordinates of contour seed pixels for
        tracing.

    Notes
    -----
    Can produce a large number of thin "halo" objects surrounding the objects
    with higher scores. These can be removed by filtering object width in the
    resulting label image.

    Returns
    -------
    Label : array_like
        A uint32 label image.

    See Also
    --------
    ScoreContours, TraceContours, MinimumModel

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    from skimage.draw import polygon

    # initialize label image
    Label = np.zeros(Shape, dtype=np.dtype('uint32'))

    # sort contours by scores
    Order = np.argsort(Scores)

    # loop over sorted contours, from least to most prominently scores
    for i in np.arange(len(cXs)):

        # get limits on contour
        xMin = np.min(cXs[Order[i]])
        xMax = np.max(cXs[Order[i]])
        yMin = np.min(cYs[Order[i]])
        yMax = np.max(cYs[Order[i]])

        # extract portion of existing label image
        T = Label[yMin:yMax + 1, xMin:xMax + 1]

        # generate mask for object 'Order[i]' from polygon
        Mask = polygon(cYs[Order[i]] - yMin, cXs[Order[i]] - xMin, T.shape)

        # replace non-zero areas with value 'i'
        T[Mask] = i

    return Label


def split_concavities(Label, MinDepth=4, MinConcavity=np.inf):  # noqa: C901
    """Performs splitting of objects in a label image using geometric scoring
    of concavities. Attempts to perform splits at narrow regions that are
    perpendicular to the object's convex hull boundaries.

    Parameters
    ----------
    Label : array_like
        A uint32 label image.
    MinDepth : float
        Minimum depth of concavities to consider during geometric splitting.
        Default value = 2.
    MinConcavity : float
        Minimum concavity score to consider when performing for geometric
        splitting. Default value = np.inf.

    Notes
    -----
    Can produce a large number of thin "halo" objects surrounding the objects
    with higher scores. These can be removed by filtering object width in the
    resulting label image.

    Returns
    -------
    Label : array_like
        A uint32 label image.

    See Also
    --------
    label_contours, min_model

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    import scipy
    import scipy.ndimage as ndi
    import skimage.morphology as mo

    # use shape profiles to split objects with concavities
    # copy input label image
    Convex = Label.copy()

    # condense label image
    if np.unique(Convex).size - 1 != Convex.max():
        Convex = label.condense(Convex)

    # get locations of objects in initial image
    Locations = scipy.ndimage.find_objects(Convex)

    # initialize number of labeled objects and record initial count
    Total = Label.max()

    # initialize loop counter
    i = 1

    while i <= Total:

        # get object window from label image
        if i < len(Locations):
            W = Convex[Locations[i - 1]]
        else:
            Locations = scipy.ndimage.find_objects(Convex)
            W = Convex[Locations[i - 1]]

        # embed masked object in padded boolean array
        Mask = np.zeros((W.shape[0] + 2, W.shape[1] + 2), dtype=bool)
        Mask[1:-1, 1:-1] = W == i

        # generate convex hull of object
        Hull = mo.convex_hull_image(Mask)

        # generate boundary coordinates, trim duplicate point
        X, Y = label.trace_object_boundaries(Mask, conn=8)
        X = np.array(X[0][:-1], dtype=np.uint32)
        Y = np.array(Y[0][:-1], dtype=np.uint32)

        # calculate distance transform of object boundary pixels to convex hull
        Distance = ndi.distance_transform_edt(Hull)
        D = Distance[Y, X] - 1

        # generate linear index of positions
        Linear = np.arange(X.size)

        # rotate boundary counter-clockwise until start position is on hull
        while D[0] != 0:
            X = np.roll(X, -1)
            Y = np.roll(Y, -1)
            D = np.roll(D, -1)
            Linear = np.roll(Linear, -1)

        # find runs of concave pixels with length > 1
        Concave = (D > 0).astype(int)
        Start = np.where((Concave[1:] - Concave[0:-1]) == 1)[0]
        Stop = np.where((Concave[1:] - Concave[0:-1]) == -1)[0] + 1
        if Stop.size == Start.size - 1:
            Stop = np.append(Stop, 0)

        # extract depth profiles, indices, distances for each run
        iX = []
        iY = []
        Depths = []
        Length = np.zeros(Start.size)
        MaxDepth = np.zeros(Start.size)
        for j in np.arange(Start.size):
            if Start[j] < Stop[j]:
                iX.append(X[Start[j]:Stop[j] + 1])
                iY.append(Y[Start[j]:Stop[j] + 1])
                Depths.append(D[Start[j]:Stop[j] + 1])
            else:  # run terminates at beginning of sequence
                iX.append(np.append(X[Start[j]:], X[0]))
                iY.append(np.append(Y[Start[j]:], Y[0]))
                Depths.append(np.append(D[Start[j]:], D[0]))
            Length[j] = iX[j].size
            MaxDepth[j] = np.max(Depths[j])

        # filter based on concave contour length and max depth
        Keep = np.where((Length > 1) & (MaxDepth >= MinDepth))[0]
        Start = Start[Keep]
        Stop = Stop[Keep]
        iX = [iX[Ind].astype(dtype=float) for Ind in Keep]
        iY = [iY[Ind].astype(dtype=float) for Ind in Keep]
        Depths = [Depths[Ind] for Ind in Keep]

        # attempt cutting if more than 1 sequence is found
        if Start.size > 1:

            # initialize containers to hold cut scores, optimal cut locations
            Scores = np.inf * np.ones((Start.size, Start.size))
            Xcut1 = np.zeros((Start.size, Start.size), dtype=np.uint32)
            Ycut1 = np.zeros((Start.size, Start.size), dtype=np.uint32)
            Xcut2 = np.zeros((Start.size, Start.size), dtype=np.uint32)
            Ycut2 = np.zeros((Start.size, Start.size), dtype=np.uint32)

            # compare candidates pairwise between all runs and score
            for j in np.arange(Start.size):

                # get list of 'j' candidates that pass depth threshold
                jCandidates = np.where(Depths[j] >= MinDepth)[0]

                for k in np.arange(j + 1, Start.size):

                    # get list of 'k' candidates that pass depth threshold
                    kCandidates = np.where(Depths[k] >= MinDepth)[0]

                    # initialize minimum score and cut locations
                    minScore = np.inf
                    minj = -1
                    mink = -1

                    # loop over each coordinate pair for concavities j,k
                    for a in np.arange(jCandidates.size):
                        for b in np.arange(kCandidates.size):

                            # calculate length score
                            Ls = length_score(iX[j][jCandidates[a]],
                                              iY[j][jCandidates[a]],
                                              iX[k][kCandidates[b]],
                                              iY[k][kCandidates[b]],
                                              Depths[j][jCandidates[a]],
                                              Depths[k][kCandidates[b]])

                            # calculate angle score
                            As = angle_score(iX[j][0], iY[j][0],
                                             iX[j][-1], iY[j][-1],
                                             iX[k][0], iY[k][0],
                                             iX[k][-1], iY[k][-1],
                                             iX[j][jCandidates[a]],
                                             iY[j][jCandidates[a]],
                                             iX[k][kCandidates[b]],
                                             iY[k][kCandidates[b]])

                            # combine scores
                            Score = (Ls + As) / 2

                            # replace if improvement
                            if Score < minScore:
                                minScore = Score
                                Scores[j, k] = minScore
                                minj = jCandidates[a]
                                mink = kCandidates[b]

                    # record best cut location
                    Xcut1[j, k] = iX[j][minj]
                    Ycut1[j, k] = iY[j][minj]
                    Xcut2[j, k] = iX[k][mink]
                    Ycut2[j, k] = iY[k][mink]

            # pick the best scoring candidates and cut if needed
            ArgMin = np.unravel_index(Scores.argmin(), Scores.shape)
            if Scores[ArgMin[0], ArgMin[1]] <= MinConcavity:

                # perform cut
                SplitMask = cut(Mask,
                                Xcut1[ArgMin[0], ArgMin[1]].astype(float),
                                Ycut1[ArgMin[0], ArgMin[1]].astype(float),
                                Xcut2[ArgMin[0], ArgMin[1]].astype(float),
                                Ycut2[ArgMin[0], ArgMin[1]].astype(float))

                # re-label cut image
                SplitLabel = scipy.ndimage.label(SplitMask)[0]

                # increment object count, and label new object at end
                SplitLabel[SplitLabel > 1] = Total + 1
                Total += 1

                # label object '1' with current object value
                SplitLabel[SplitLabel == 1] = i

                # trim padding from corrected label image
                Mask = Mask[1:-1, 1:-1]
                SplitLabel = SplitLabel[1:-1, 1:-1]

                # update label image
                W[Mask] = SplitLabel[Mask]

            else:  # no cut made, move to next object
                i = i + 1
        else:  # no cuts to attempt, move to next object
            i = i + 1

    return Convex


def angle_score(ax1, ay1, bx1, by1, ax2, ay2, bx2, by2, cx1, cy1, cx2, cy2):
    """Scores the angles produced by cutting line (cx1, cy1)->(cx2, cy2) given
    the convex hull segments (ax1, ay1)->(bx1, by1) and (ax2, ay2)->(bx2, by2)
    spanning the concavities. See Figure 6 in reference below for a full
    illustration.

    Returns
    -------
    Score : float
        Angle score according to equation (6) from the reference.

    See Also
    --------
    SplitConcavities

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    # calculate angle of hull at first concavity - y is inverted
    jHullAlpha = np.arctan2(ay1 - by1, bx1 - ax1)

    # calculate angle of cut at first concavity - y is inverted
    jCutAlpha = np.arctan2(cy2 - cy1, cx1 - cx2)

    # calculate angle of hull at second concavity - y is inverted
    kHullAlpha = np.arctan2(ay2 - by2, bx2 - ax2)

    # calculate angle of cut at second concavity - y is inverted
    kCutAlpha = np.arctan2(cy1 - cy2, cx2 - cx1)

    # calculate angle score
    Score = (np.abs(np.pi / 2 - (jCutAlpha - jHullAlpha)) +
             np.abs(np.pi / 2 - (kCutAlpha - kHullAlpha))) / np.pi

    return Score


def length_score(x1, y1, x2, y2, d1, d2):
    """Scores the length of the cutting line (x1, y1)->(x2, y2) made at a
    concavity depth of d1 and d2.

    Returns
    -------
    Score : float
        Angle score according to equation (5) from the reference.

    See Also
    --------
    SplitConcavities

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    # calculate length of cut
    r = ((x1 - x2)**2 + (y1 - y2)**2) ** 0.5

    # normalize by total span across convex hull
    LengthScore = r / (r + d1 + d2)

    return LengthScore


def cut(Mask, x1, y1, x2, y2):
    """Performs a cut across a binary mask, zeroing pixels that round to
    positions on the line (x1, y1)->(x2, y2).

    Returns
    -------
    Cut : array_like
        A version of input Mask modified by cutting the line (x1, y1)->(x2, y2)

    See Also
    --------
    SplitConcavities

    References
    ----------
    .. [#] S. Weinert et al "Detection and Segmentation of Cell Nuclei in
    Virtual Microscopy Images: A Minimum-Model Approach" in Nature Scientific
    Reports,vol.2,no.503, doi:10.1038/srep00503, 2012.

    """
    # copy input
    Cut = Mask.copy()

    # calculate angle of line
    if x1 < x2:
        theta = np.arctan2(y2 - y1, x2 - x1)
    else:
        theta = np.arctan2(y1 - y2, x1 - x2)

    # define line length
    length = ((x1 - x2)**2 + (y1 - y2)**2)**0.5

    # define points along x-axis
    x = np.arange(-1, length + 1, 0.1)
    y = np.zeros(x.shape)

    # rotate
    R = np.array([[np.cos(theta), -np.sin(theta)],
                  [np.sin(theta), np.cos(theta)]])
    rotated = R.dot(np.vstack((x, y)))

    # translate
    if x1 < x2:
        xr = rotated[0, :] + x1
        yr = rotated[1, :] + y1
    else:
        xr = rotated[0, :] + x2
        yr = rotated[1, :] + y2

    # take floor of coordinates
    xr = np.round(xr).astype(np.uint32)
    yr = np.round(yr).astype(np.uint32)

    # remove any pairs containing negative numbers to avoid access violations
    negative = (xr < 0) | (yr < 0)
    xr = np.delete(xr, negative)
    yr = np.delete(yr, negative)

    # zero out these coordinates
    Cut[yr, xr] = False

    return Cut



import json
from pathlib import Path

import numpy

import histomicstk
from histomicstk.cli import utils
from histomicstk.cli.utils import CLIArgumentParser
from histomicstk.preprocessing.color_deconvolution import \
    rgb_separate_stains_macenko_pca


def main(origargs):
    args = utils.splitArgs(origargs)
    args.macenko.I_0 = numpy.array(args.macenko.I_0)

    utils.create_dask_client(args.dask)
    sample = utils.sample_pixels(args.sample)
    stain_matrix = rgb_separate_stains_macenko_pca(sample.T, **vars(args.macenko))

    # record metadata
    annotation = {
        'name': 'SeperateStainsMacenkoPCA',
        'attributes': {
            'params': vars(origargs),
            'cli': Path(__file__).stem,
            'version': histomicstk.__version__,
        },
    }
    for i, stain in enumerate(stain_matrix.T):
        annotation['attributes'][f'stainColor_{i + 1}'] = stain.tolist()

    with open(args.outputAnnotationFile, 'w') as annotation_file:
        json.dump(annotation, annotation_file, sort_keys=False)


if __name__ == '__main__':
    main(CLIArgumentParser().parse_args())



"""
This package contains functions to computing a variety of image-based features
that quantify the appearance and/or morphology of an objects/regions in the
image. These are needed for classifying objects (e.g. nuclei) and
regions (e.g. tissues) found in histopathology images.
"""

# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .compute_fsd_features import compute_fsd_features
from .compute_global_cell_graph_features import \
    compute_global_cell_graph_features
from .compute_gradient_features import compute_gradient_features
from .compute_haralick_features import compute_haralick_features
from .compute_intensity_features import compute_intensity_features
from .compute_morphometry_features import compute_morphometry_features
from .compute_nuclei_features import compute_nuclei_features
from .graycomatrixext import graycomatrixext

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'compute_fsd_features',
    'compute_global_cell_graph_features',
    'compute_gradient_features',
    'compute_haralick_features',
    'compute_intensity_features',
    'compute_morphometry_features',
    'compute_nuclei_features',
    'graycomatrixext',
)



"""
Created on Tue Oct 22 02:37:52 2019.

@author: mtageld
"""
import numpy as np
from PIL import Image

from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    get_image_from_htk_response
from histomicstk.annotations_and_masks.masks_to_annotations_handler import (
    get_annotation_documents_from_contours, get_contours_from_mask)
from histomicstk.features.compute_intensity_features import \
    compute_intensity_features
from histomicstk.preprocessing.color_conversion import (lab_mean_std,
                                                        rgb_to_hsi, rgb_to_lab)
from histomicstk.preprocessing.color_deconvolution import (
    _reorder_stains, color_deconvolution_routine,
    rgb_separate_stains_macenko_pca)
from histomicstk.preprocessing.color_normalization import (
    deconvolution_based_normalization, reinhard)
from histomicstk.saliency.tissue_detection import (_get_largest_regions,
                                                   get_slide_thumbnail,
                                                   get_tissue_mask,
                                                   threshold_multichannel)
from histomicstk.utils.general_utils import Base_HTK_Class

Image.MAX_IMAGE_PIXELS = None


class CDT_single_tissue_piece:
    """Detect various regions in a single tissue piece (internal)."""

    def __init__(self, cdt, tissue_mask, monitorPrefix=''):
        """Detect whitespace, saliency, etc in one tissue piece (Internal).

        Arguments:
        ---------
        cdt : object
            Cellularity_detector_thresholding instance
        tissue_mask : np array
            (m x n) mask of the tissue piece at cdt.MAG magnification
        monitorPrefix : str
            Text to prepend to printed statements

        """
        self.cdt = cdt
        self.tissue_mask = 0 + tissue_mask
        self.monitorPrefix = monitorPrefix

    def run(self):
        """Get cellularity and optionally visualize on DSA."""
        self.restrict_mask_to_single_tissue_piece()
        self.cdt._print2('%s: set_tissue_rgb()' % self.monitorPrefix)
        self.set_tissue_rgb()
        self.cdt._print2('%s: initialize_labeled_mask()' % self.monitorPrefix)
        self.initialize_labeled_mask()
        self.cdt._print2(
            '%s: assign_components_by_thresholding()' % self.monitorPrefix)
        self.assign_components_by_thresholding()
        self.cdt._print2(
            '%s: color_normalize_unspecified_components()'
            % self.monitorPrefix)
        self.color_normalize_unspecified_components()
        self.cdt._print2(
            '%s: find_potentially_cellular_regions()' % self.monitorPrefix)
        self.find_potentially_cellular_regions()
        self.cdt._print2(
            '%s: find_top_cellular_regions()' % self.monitorPrefix)
        self.find_top_cellular_regions()
        if self.cdt.visualize:
            self.cdt._print2('%s: visualize_results()' % self.monitorPrefix)
            self.visualize_results()

    def restrict_mask_to_single_tissue_piece(self):
        """Only keep relevant part of slide mask."""
        # find coordinates at scan magnification
        tloc = np.argwhere(self.tissue_mask)
        F = self.cdt.slide_info['F_tissue']
        self.ymin, self.xmin = (int(j) for j in np.min(tloc, axis=0) * F)
        self.ymax, self.xmax = (int(j) for j in np.max(tloc, axis=0) * F)
        self.tissue_mask = self.tissue_mask[
            int(self.ymin / F): int(self.ymax / F),
            int(self.xmin / F): int(self.xmax / F)]

    def set_tissue_rgb(self):
        """Load RGB from server for single tissue piece."""
        # load RGB for this tissue piece at saliency magnification
        getStr = '/item/%s/tiles/region?left=%d&right=%d&top=%d&bottom=%d&encoding=PNG' % (
            self.cdt.slide_id, self.xmin, self.xmax, self.ymin, self.ymax,
        ) + '&magnification=%d' % self.cdt.MAG
        resp = self.cdt.gc.get(getStr, jsonResp=False)
        self.tissue_rgb = get_image_from_htk_response(resp)

    def initialize_labeled_mask(self):
        """Initialize labeled components mask."""
        from skimage.transform import resize

        # resize tissue mask to target mag
        self.labeled = resize(
            self.tissue_mask, output_shape=self.tissue_rgb.shape[:2],
            order=0, preserve_range=True, anti_aliasing=False)
        self.labeled[self.labeled > 0] = self.cdt.GTcodes.loc[
            'not_specified', 'GT_code']

    def assign_components_by_thresholding(self):
        """Get components by thresholding in HSI and LAB spaces."""
        # get HSI and LAB images
        self.cdt._print2(
            '%s: -- get HSI and LAB images ...' % self.monitorPrefix)
        tissue_hsi = rgb_to_hsi(self.tissue_rgb)
        tissue_lab = rgb_to_lab(self.tissue_rgb)

        # extract components using HSI/LAB thresholds

        hsi_components = self.cdt.hsi_thresholds.keys()
        lab_components = self.cdt.lab_thresholds.keys()

        for component in self.cdt.ordered_components:

            self.cdt._print2('%s: -- thresholding %s ...' % (
                self.monitorPrefix, component))

            if component in hsi_components:
                lab, _ = threshold_multichannel(
                    tissue_hsi,
                    channels=['hue', 'saturation', 'intensity'],
                    thresholds=self.cdt.hsi_thresholds[component],
                    just_threshold=False,
                    get_tissue_mask_kwargs=self.cdt.get_tissue_mask_kwargs2)
            elif component in lab_components:
                lab, _ = threshold_multichannel(
                    tissue_lab,
                    channels=['l', 'a', 'b'],
                    thresholds=self.cdt.lab_thresholds[component],
                    just_threshold=True,
                    get_tissue_mask_kwargs=self.cdt.get_tissue_mask_kwargs2)
            else:
                msg = 'Unknown component name.'
                raise ValueError(msg)

            lab[self.labeled == 0] = 0  # restrict to tissue mask
            self.labeled[lab > 0] = self.cdt.GTcodes.loc[component, 'GT_code']

        # This deals with holes in tissue
        self.labeled[self.labeled == 0] = self.cdt.GTcodes.loc[
            'outside_tissue', 'GT_code']

    def color_normalize_unspecified_components(self):
        """Color normalize "true" tissue components."""
        if self.cdt.color_normalization_method == 'reinhard':
            self.cdt._print2(
                '%s: -- reinhard normalization ...' % self.monitorPrefix)
            self.tissue_rgb = reinhard(
                self.tissue_rgb,
                target_mu=self.cdt.target_stats_reinhard['mu'],
                target_sigma=self.cdt.target_stats_reinhard['sigma'],
                mask_out=self.labeled != self.cdt.GTcodes
                .loc['not_specified', 'GT_code'])

        elif self.cdt.color_normalization_method == 'macenko_pca':
            self.cdt._print2(
                '%s: -- macenko normalization ...' % self.monitorPrefix)
            self.tissue_rgb = deconvolution_based_normalization(
                self.tissue_rgb, W_target=self.cdt.target_W_macenko,
                mask_out=self.labeled != self.cdt.GTcodes
                .loc['not_specified', 'GT_code'],
                stain_unmixing_routine_params=self.
                cdt.stain_unmixing_routine_params)
        else:
            self.cdt._print2('%s: -- No normalization!' % self.monitorPrefix)

    def find_potentially_cellular_regions(self):
        """Find regions that are potentially cellular."""
        from scipy import ndimage
        from skimage.filters import gaussian

        mask_out = self.labeled != self.cdt.GTcodes.loc[
            'not_specified', 'GT_code']

        # deconvolvve to ge hematoxylin channel (cellular areas)
        # hematoxylin channel return shows MINIMA so we invert
        self.tissue_htx, _, _ = color_deconvolution_routine(
            self.tissue_rgb, mask_out=mask_out,
            **self.cdt.stain_unmixing_routine_params)
        self.tissue_htx = 255 - self.tissue_htx[..., 0]

        # get cellular regions by threshold HTX stain channel
        self.maybe_cellular, _ = get_tissue_mask(
            self.tissue_htx.copy(), deconvolve_first=False,
            n_thresholding_steps=1, sigma=self.cdt.cellular_step1_sigma,
            min_size=self.cdt.cellular_step1_min_size)

        # Second, low-pass filter to dilate and smooth a bit
        self.maybe_cellular = gaussian(
            0 + (self.maybe_cellular > 0), sigma=self.cdt.cellular_step2_sigma,
            output=None, mode='nearest', preserve_range=True)

        # find connected components
        self.maybe_cellular, _ = ndimage.label(self.maybe_cellular)

        # restrict cellular regions to not-otherwise-specified
        self.maybe_cellular[mask_out] = 0

        # assign to mask
        self.labeled[self.maybe_cellular > 0] = self.cdt.GTcodes.loc[
            'maybe_cellular', 'GT_code']

    def find_top_cellular_regions(self):
        """Keep largest and most cellular regions."""
        # keep only largest n regions regions
        top_cellular_mask = _get_largest_regions(
            self.maybe_cellular, top_n=self.cdt.cellular_largest_n)
        top_cellular = self.maybe_cellular.copy()
        top_cellular[top_cellular_mask == 0] = 0

        # get intensity features of hematoxylin channel for each region
        intensity_feats = compute_intensity_features(
            im_label=top_cellular, im_intensity=self.tissue_htx,
            feature_list=['Intensity.Mean'])
        unique = np.unique(top_cellular[top_cellular > 0])
        intensity_feats.index = unique

        # get top n brightest regions from the largest areas
        intensity_feats.sort_values('Intensity.Mean', axis=0, inplace=True)
        discard = np.array(intensity_feats.index[:-self.cdt.cellular_top_n])
        discard = np.in1d(top_cellular, discard).reshape(top_cellular.shape)
        top_cellular[discard] = 0

        # integrate into labeled mask
        self.labeled[top_cellular > 0] = self.cdt.GTcodes.loc[
            'top_cellular', 'GT_code']

    def visualize_results(self):
        """Visualize results in DSA."""
        # get contours
        contours_df = get_contours_from_mask(
            MASK=self.labeled, GTCodes_df=self.cdt.GTcodes.copy(),
            groups_to_get=self.cdt.groups_to_visualize,
            get_roi_contour=self.cdt.get_roi_contour, roi_group='roi',
            background_group='not_specified',
            discard_nonenclosed_background=True,
            MIN_SIZE=15, MAX_SIZE=None,
            verbose=self.cdt.verbose == 3,
            monitorPrefix=self.monitorPrefix + ': -- contours')

        # get annotation docs
        annprops = {
            'F': self.cdt.slide_info['magnification'] / self.cdt.MAG,
            'X_OFFSET': self.xmin,
            'Y_OFFSET': self.ymin,
            'opacity': self.cdt.opacity,
            'lineWidth': self.cdt.lineWidth,
        }
        annotation_docs = get_annotation_documents_from_contours(
            contours_df.copy(), separate_docs_by_group=True,
            docnamePrefix=self.cdt.docnameprefix,
            annprops=annprops, verbose=self.cdt.verbose == 3,
            monitorPrefix=self.monitorPrefix + ': -- annotation docs')

        # post annotations to slide
        for doc in annotation_docs:
            _ = self.cdt.gc.post(
                '/annotation?itemId=' + self.cdt.slide_id, json=doc)


class Cellularity_detector_thresholding(Base_HTK_Class):
    """Detect cellular regions in a slide using thresholding.

    This uses a thresholding and stain unmixing based pipeline
    to detect highly-cellular regions in a slide. The run()
    method of the CDT_single_tissue_piece() class has the key
    steps of the pipeline. In summary, here are the steps
    involved...

    1. Detect tissue from background using the RGB slide
    thumbnail. Each "tissue piece" is analysed independently
    from here onwards. The tissue_detection modeule is used
    for this step. A high sensitivity, low specificity setting
    is used here.

    2. Fetch the RGB image of tissue at target magnification. A
    low magnification (default is 3.0) is used and is sufficient.

    3. The image is converted to HSI and LAB spaces. Thresholding
    is performed to detect various non-salient components that
    often throw-off the color normalization and deconvolution
    algorithms. Thresholding includes both minimum and maximum
    values. The user can set whichever thresholds of components
    they would like. The development of this workflow was focused
    on breast cancer so the thresholded components by default
    are whote space (or adipose tissue), dark blue/green blotches
    (sharpie, inking at margin, etc), and blood. Whitespace
    is obtained by thresholding the saturation and intensity,
    while other components are obtained by thresholding LAB.

    4. Now that we know where "actual" tissue is, we do a MASKED
    color normalization to a prespecified standard. The masking
    ensures the normalization routine is not thrown off by non-
    tissue components.

    5. Perform masked stain unmixing/deconvolution to obtain the
    hematoxylin stain channel.

    6. Smooth and threshold the hematoxylin channel. Then
    perform connected component analysis to find contiguous
    potentially-cellular regions.

    7. Keep the n largest potentially-cellular regions. Then
    from those large regions, keep the m brightest regions
    (using hematoxylin channel brightness) as the final
    salient/cellular regions.

    """

    def __init__(self, gc, slide_id, GTcodes, **kwargs):
        """Init Cellularity_Detector_Superpixels object.

        Arguments:
        ---------
        gc : object
            girder client object

        slide_id : str
            girder ID of slide

        GTcodes : pandas Dataframe
            the ground truth codes and information dataframe.
            WARNING: Modified inside this method so pass a copy.
            This is a dataframe that is indexed by the annotation group name
            and has the following columns...

            group: str
                group name of annotation, eg. mostly_tumor
            overlay_order: int
                how early to place the annotation in the
                mask. Larger values means this annotation group is overlaid
                last and overwrites whatever overlaps it.
            GT_code: int
                desired ground truth code (in the mask).
                Pixels of this value belong to corresponding group (class)
            is_roi: bool
                whether this group encodes an ROI
            is_background_class: bool
                whether this group is the default fill value inside the ROI.
                For example, you may decide that any pixel inside the ROI
                is considered stroma.
            color: str
                rgb format. eg. rgb(255,0,0)

            The following indexes must be present...
            outside_tissue, not_specified, maybe_cellular, top_cellular

        verbose : int
            0 - Do not print to screen
            1 - Print only key messages
            2 - Print everything to screen
            3 - print everything including from inner functions

        monitorPrefix : str
            text to prepend to printed statements

        logging_savepath : str or None
            where to save run logs

        suppress_warnings : bool
            whether to suppress warnings

        MAG : float
            magnification at which to detect cellularity

        color_normalization_method : str
            Must be in ['reinhard', 'macenko_pca', 'none']

        target_W_macenko : np array
            3 by 3 stain matrix for macenko normalization
            obtained using rgb_separate_stains_macenko_pca()
            and reordered such that hematoxylin and eosin are
            the first and second channels, respectively.

        target_stats_reinhard : dict
            must contains the keys mu and sigma. Mean and sigma
            of target image in LAB space for reinhard normalization.

        get_tissue_mask_kwargs : dict
            kwargs for the get_tissue_mask() method. This is used
            to detect tissue from the slide thumbnail.

        keep_components : list
            list of strings. Names of components to exclude by
            HSI thresholding. These much be present in the index
            of the GTcodes dataframe

        get_tissue_mask_kwargs2 : dict
            kwargs for get_tissue_mask() used for iterative smoothing
            and thresholding the component masks after initial
            thresholding using the user-defined HSI/LAB thresholds.

        hsi_thresholds : dict
            each entry is a dict containing the keys hue, saturation
            and intensity. Each of these is in turn also a dict
            containing the keys min and max. See default value below
            for an example.

        lab_thresholds : dict
            each entry is a dict containing the keys l, a, and b.
            Each of these is in turn also a dict containing the keys
            min and max. See default value below for an example.

        stain_unmixing_routine_params : dict
            kwargs passed as the stain_unmixing_routine_params
            argument to the deconvolution_based_normalization method

        cellular_step1_sigma : float
            sigma of gaussian smoothing for first cellularity step

        cellular_step1_min_size : int
            minimum contiguous size for first cellularity step

        cellular_step2_sigma : float
            sigma of gaussian smoothing for second cellularity step

        cellular_largest_n : int
            Number of large continugous cellular regions to keep

        cellular_top_n : int
            Number of final "top" cellular regions to keep

        visualize : bool
            whether to visualize results in DSA

        opacity : float
            opacity of superpixel polygons when posted to DSA.
            0 (no opacity) is more efficient to render.

        lineWidth : float
            width of line when displaying region boundaries.

        docnameprefix : str
            prefix to add to annotation document name

        groups_to_visualize : list
            which groups to visualize

        get_roi_contour : bool
            whether to get the contour of the roi

        """
        default_attr = {

            # The following are already assigned defaults by Base_HTK_Class
            # 'verbose': 1,
            # 'monitorPrefix': "",
            # 'logging_savepath': None,
            # 'suppress_warnings': False,

            'MAG': 3.0,

            # Must be in ['reinhard', 'macenko_pca', 'none']
            'color_normalization_method': 'macenko_pca',

            # TCGA-A2-A3XS-DX1_xmin21421_ymin37486_.png, Amgad et al, 2019)
            # is used as the target image for reinhard & macenko normalization

            # for macenco (obtained using rgb_separate_stains_macenko_pca()
            # and using reordered such that columns are the order:
            # Hamtoxylin, Eosin, Null
            'target_W_macenko': np.array([
                [0.5807549, 0.08314027, 0.08213795],
                [0.71681094, 0.90081588, 0.41999816],
                [0.38588316, 0.42616716, -0.90380025],
            ]),

            # TCGA-A2-A3XS-DX1_xmin21421_ymin37486_.png, Amgad et al, 2019)
            # Reinhard color norm. standard
            'target_stats_reinhard': {
                'mu': np.array([8.74108109, -0.12440419, 0.0444982]),
                'sigma': np.array([0.6135447, 0.10989545, 0.0286032]),
            },

            # kwargs for getting masks for all tissue pieces (thumbnail)
            'get_tissue_mask_kwargs': {
                'deconvolve_first': True, 'n_thresholding_steps': 1,
                'sigma': 1.5, 'min_size': 500,
            },

            # components to extract by HSI thresholding
            'keep_components': ['blue_sharpie', 'blood', 'whitespace'],

            # kwargs for getting components masks
            'get_tissue_mask_kwargs2': {
                'deconvolve_first': False, 'n_thresholding_steps': 1,
                'sigma': 5.0, 'min_size': 50,
            },

            # min/max thresholds for HSI and LAB
            'hsi_thresholds': {
                'whitespace': {
                    'hue': {'min': 0, 'max': 1.0},
                    'saturation': {'min': 0, 'max': 0.2},
                    'intensity': {'min': 220, 'max': 255},
                },
            },
            'lab_thresholds': {
                'blue_sharpie': {
                    'l': {'min': -1000, 'max': 1000},
                    'a': {'min': -1000, 'max': 1000},
                    'b': {'min': -1000, 'max': -0.02},
                },
                'blood': {
                    'l': {'min': -1000, 'max': 1000},
                    'a': {'min': 0.02, 'max': 1000},
                    'b': {'min': -1000, 'max': 1000},
                },
            },

            # for stain unmixing to deconvolove and/or color normalize
            'stain_unmixing_routine_params': {
                'stains': ['hematoxylin', 'eosin'],
                'stain_unmixing_method': 'macenko_pca',
            },

            # params for getting cellular regions
            'cellular_step1_sigma': 0.,
            'cellular_step1_min_size': 100,
            'cellular_step2_sigma': 1.5,
            'cellular_largest_n': 5,
            'cellular_top_n': 2,

            # visualization params
            'visualize': True,
            'opacity': 0,
            'lineWidth': 3.0,
            'docnameprefix': 'cdt',
            'groups_to_visualize': None,  # everything
            'get_roi_contour': True,
        }
        default_attr.update(kwargs)
        super().__init__(
            default_attr=default_attr)

        self.color_normalization_method = \
            self.color_normalization_method.lower()
        assert self.color_normalization_method in [
            'reinhard', 'macenko_pca', 'none']

        # set attribs
        self.gc = gc
        self.slide_id = slide_id
        self.GTcodes = GTcodes

        self.fix_GTcodes()

    def fix_GTcodes(self):
        """Fix self.GTcodes (important!)."""
        # validate
        self.GTcodes.index = self.GTcodes.loc[:, 'group']
        necessary_indexes = self.keep_components + [
            'outside_tissue', 'not_specified',
            'maybe_cellular', 'top_cellular']
        assert all(j in list(self.GTcodes.index) for j in necessary_indexes)

        # Make sure the first things laid out are the "background" components
        min_val = np.min(self.GTcodes.loc[:, 'overlay_order'])
        self.GTcodes.loc['outside_tissue', 'overlay_order'] = min_val - 2
        self.GTcodes.loc['not_specified', 'overlay_order'] = min_val - 1

        # reorder in overlay order (important)
        self.GTcodes.sort_values('overlay_order', axis=0, inplace=True)
        self.ordered_components = list(self.GTcodes.loc[:, 'group'])

        # only keep relevant components (for HSI/LAB thresholding)
        for c in self.ordered_components[:]:
            if c not in self.keep_components:
                self.ordered_components.remove(c)

    def run(self):
        """Run full pipeline to detect cellular regions."""
        # get mask, each unique value is a single tissue piece
        self._print1(
            '%s: set_slide_info_and_get_tissue_mask()' % self.monitorPrefix)
        labeled = self.set_slide_info_and_get_tissue_mask()

        # Go through tissue pieces and do run sequence
        unique_tvals = list(set(np.unique(labeled)) - {0})
        tissue_pieces = [None for _ in range(len(unique_tvals))]
        for idx, tval in enumerate(unique_tvals):
            monitorPrefix = '%s: Tissue piece %d of %d' % (
                self.monitorPrefix, idx + 1, len(unique_tvals))
            self._print1(monitorPrefix)
            tissue_pieces[idx] = CDT_single_tissue_piece(
                self, tissue_mask=labeled == tval, monitorPrefix=monitorPrefix)
            tissue_pieces[idx].run()
            # delete unnecessary attributes
            del (
                tissue_pieces[idx].tissue_rgb,  # too much space
                tissue_pieces[idx].tissue_mask,  # already part of labeled
                tissue_pieces[idx].maybe_cellular,  # already part of labeled
                tissue_pieces[idx].tissue_htx,  # unnecessary
            )

        return tissue_pieces

    def set_color_normalization_target(
            self, ref_image_path, color_normalization_method='macenko_pca'):
        """Set color normalization values to use from target image.

        Arguments:
        ---------
        ref_image_path, str
        >    path to target (reference) image
        color_normalization_method, str
        >    color normalization method to use. Currently, only
        >    reinhard and macenko_pca are accepted.

        """
        from imageio import imread

        # read input image
        ref_im = np.array(imread(ref_image_path, pilmode='RGB'))

        # assign target values

        color_normalization_method = color_normalization_method.lower()

        if color_normalization_method == 'reinhard':
            mu, sigma = lab_mean_std(ref_im)
            self.target_stats_reinhard['mu'] = mu
            self.target_stats_reinhard['sigma'] = sigma

        elif color_normalization_method == 'macenko_pca':
            self.target_W_macenko = _reorder_stains(
                rgb_separate_stains_macenko_pca(ref_im, I_0=None),
                stains=['hematoxylin', 'eosin'])
        else:
            raise ValueError(
                'Unknown color_normalization_method: %s' %
                (color_normalization_method))

        self.color_normalization_method = color_normalization_method

    def set_slide_info_and_get_tissue_mask(self):
        """Set self.slide_info dict and self.labeled tissue mask."""
        # This is a persistent dict to store information about slide
        self.slide_info = self.gc.get('item/%s/tiles' % self.slide_id)

        # get tissue mask
        thumbnail_rgb = get_slide_thumbnail(self.gc, self.slide_id)

        # get labeled tissue mask -- each unique value is one tissue piece
        labeled, _ = get_tissue_mask(
            thumbnail_rgb, **self.get_tissue_mask_kwargs)

        if len(np.unique(labeled)) < 2:
            msg = 'No tissue detected!'
            raise ValueError(msg)

        # Find size relative to WSI
        self.slide_info['F_tissue'] = self.slide_info[
            'sizeX'] / labeled.shape[1]

        return labeled



"""
Created on Wed Sep 18 00:06:28 2019.

@author: mtageld
"""
import os
import shutil
import sys
import tempfile

import numpy as np
import pytest
from pandas import read_csv

from histomicstk.annotations_and_masks.annotation_and_mask_utils import \
    delete_annotations_in_slide
from histomicstk.saliency.cellularity_detection_superpixels import \
    Cellularity_detector_superpixels
from histomicstk.saliency.cellularity_detection_thresholding import \
    Cellularity_detector_thresholding
from histomicstk.saliency.tissue_detection import (
    get_slide_thumbnail, get_tissue_boundary_annotation_documents,
    get_tissue_mask)

thisDir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, os.path.join(thisDir, '../../../tests'))
# import htk_test_utilities as utilities  # noqa
from htk_test_utilities import getTestFilePath, girderClient  # noqa


class Cfg:
    def __init__(self):
        self.gc = None
        self.iteminfo = None
        self.labeled = None
        self.GTcodes = None
        self.logging_savepath_cdt = tempfile.mkdtemp()
        self.logging_savepath_cds = tempfile.mkdtemp()


cfg = Cfg()


@pytest.mark.usefixtures('girderClient')  # noqa
def test_prep(girderClient):  # noqa
    cfg.gc = girderClient
    cfg.iteminfo = cfg.gc.get('/item', parameters={'text': 'TCGA-A2-A0YE-01Z-00-DX1'})[0]
    cfg.GTcodes = read_csv(getTestFilePath('saliency_GTcodes.csv'))


class TestTissueDetection():
    """Test methods for getting ROI mask from annotations."""

    def test_get_tissue_mask(self):
        """Test get_tissue_mask()."""
        thumbnail_rgb = get_slide_thumbnail(cfg.gc, cfg.iteminfo['_id'])
        cfg.labeled, mask = get_tissue_mask(
            thumbnail_rgb, deconvolve_first=True,
            n_thresholding_steps=1, sigma=1.5, min_size=30)

        assert cfg.labeled.shape == (156, 256)
        assert len(np.unique(cfg.labeled)) in (10, 11)

    def test_get_tissue_boundary_annotation_documents(self):
        """Test get_tissue_boundary_annotation_documents()."""
        annotation_docs = get_tissue_boundary_annotation_documents(
            cfg.gc, slide_id=cfg.iteminfo['_id'], labeled=cfg.labeled)

        assert 'elements' in annotation_docs[0].keys()
        assert len(annotation_docs[0]['elements']) in {9, 10}

        # test delete existing annotations in slide
        delete_annotations_in_slide(cfg.gc, cfg.iteminfo['_id'])

        # check that it posts without issues
        resps = [
            cfg.gc.post(
                '/annotation?itemId=' + cfg.iteminfo['_id'], json=doc)
            for doc in annotation_docs
        ]
        assert all('annotation' in resp for resp in resps)


class TestCellularityDetection:
    """Test methods for getting cellularity."""

    def test_cellularity_detection_thresholding(self):
        """Test Cellularity_detector_thresholding()."""
        # run cellularity detector
        cdt = Cellularity_detector_thresholding(
            cfg.gc, slide_id=cfg.iteminfo['_id'],
            GTcodes=cfg.GTcodes, MAG=1.0,
            verbose=1, monitorPrefix='test',
            logging_savepath=cfg.logging_savepath_cdt)
        tissue_pieces = cdt.run()

        # check
        assert len(tissue_pieces) == 3
        assert all(j in tissue_pieces[0].__dict__.keys() for j in
                   ('labeled', 'ymin', 'xmin', 'ymax', 'xmax'))

        # cleanup
        shutil.rmtree(cfg.logging_savepath_cdt)

    def test_cellularity_detection_superpixels(self):
        """Test Cellularity_detector_superpixels()."""
        # from the ROI in Amgad et al, 2019
        cnorm_main = {
            'mu': np.array([8.74108109, -0.12440419, 0.0444982]),
            'sigma': np.array([0.6135447, 0.10989545, 0.0286032]),
        }
        # run cellularity detector
        cds = Cellularity_detector_superpixels(
            cfg.gc, slide_id=cfg.iteminfo['_id'],
            MAG=1.0, compactness=0.1, spixel_size_baseMag=256 * 256,
            max_cellularity=40,
            visualize_spixels=True, visualize_contiguous=True,
            get_tissue_mask_kwargs={
                'deconvolve_first': False,
                'n_thresholding_steps': 2,
                'sigma': 1.5,
                'min_size': 500},
            verbose=1, monitorPrefix='test',
            logging_savepath=cfg.logging_savepath_cds)
        cds.set_color_normalization_values(
            mu=cnorm_main['mu'], sigma=cnorm_main['sigma'], what='main')
        tissue_pieces = cds.run()

        # check
        assert len(tissue_pieces) == 2
        assert all(j in tissue_pieces[0].__dict__.keys()
                   for j in ('tissue_mask', 'ymin', 'xmin', 'ymax', 'xmax',
                             'spixel_mask', 'fdata', 'cluster_props'))
        assert len(tissue_pieces[0].cluster_props) == 5

        # cleanup
        shutil.rmtree(cfg.logging_savepath_cds)



import numpy


def get_principal_components(m):
    """Take a matrix m (probably 3xN) and return the 3x3 matrix of
    "principal components".  (Actually computed with SVD)

    """
    return numpy.linalg.svd(m.astype(float), full_matrices=False)[0].astype(m.dtype)


def magnitude(m):
    """Get the magnitude of each column vector in a matrix"""
    return numpy.sqrt((m ** 2).sum(0))


def normalize(m):
    """Normalize each column vector in a matrix"""
    return m / magnitude(m)



"""
This package contains utility functions that are widely used by functions in
all other sub-packages of histomicstk
"""

from . import general_utils, girder_convenience_utils
# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .compute_tile_foreground_fraction import compute_tile_foreground_fraction
from .convert_image_to_matrix import convert_image_to_matrix
from .convert_matrix_to_image import convert_matrix_to_image
from .del2 import del2
from .eigen import eigen
from .exclude_nonfinite import exclude_nonfinite
from .fit_poisson_mixture import fit_poisson_mixture
from .gradient_diffusion import gradient_diffusion
from .hessian import hessian
from .merge_colinear import merge_colinear
from .sample_pixels import sample_pixels  # must import after SimpleMask
from .simple_mask import simple_mask

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'compute_tile_foreground_fraction',
    'convert_matrix_to_image',
    'convert_image_to_matrix',
    'del2',
    'eigen',
    'exclude_nonfinite',
    'gradient_diffusion',
    'hessian',
    'merge_colinear',
    'fit_poisson_mixture',
    'sample_pixels',
    'simple_mask',
    'general_utils',
    'girder_convenience_utils',
)



import os

from slicer_cli_web import ctk_cli_adjustment  # noqa
import slicer_cli_web.cli_list_entrypoint  # noqa I001


def main():
    slicer_cli_web.cli_list_entrypoint.CLIListEntrypoint(
        cwd=os.path.dirname(os.path.realpath(__file__)))
